"""Authentication decorators and utilities for Bedrock AgentCore SDK."""

import asyncio
import contextvars
import logging
import os
from functools import wraps
from typing import Any, Callable, Dict, List, Literal, Optional

import boto3
from botocore.exceptions import ClientError

from bedrock_agentcore.runtime import BedrockAgentCoreContext
from bedrock_agentcore.services.identity import IdentityClient, TokenPoller

logger = logging.getLogger("bedrock_agentcore.auth")
logger.setLevel("INFO")
if not logger.handlers:
    logger.addHandler(logging.StreamHandler())


def requires_access_token(
    *,
    provider_name: str,
    into: str = "access_token",
    scopes: List[str],
    on_auth_url: Optional[Callable[[str], Any]] = None,
    auth_flow: Literal["M2M", "USER_FEDERATION"],
    callback_url: Optional[str] = None,
    force_authentication: bool = False,
    token_poller: Optional[TokenPoller] = None,
    custom_state: Optional[str] = None,
    custom_parameters: Optional[Dict[str, str]] = None,
) -> Callable:
    """Decorator that fetches an OAuth2 access token before calling the decorated function.

    Args:
        provider_name: The credential provider name
        into: Parameter name to inject the token into
        scopes: OAuth2 scopes to request
        on_auth_url: Callback for handling authorization URLs
        auth_flow: Authentication flow type ("M2M" or "USER_FEDERATION")
        callback_url: OAuth2 callback URL
        force_authentication: Force re-authentication
        token_poller: Custom token poller implementation
        custom_state: A state that allows applications to verify the validity of callbacks to callback_url
        custom_parameters: A map of custom parameters to include in authorization request to the credential provider
                           Note: these parameters are in addition to standard OAuth 2.0 flow parameters

    Returns:
        Decorator function
    """

    def decorator(func: Callable) -> Callable:
        client = IdentityClient(_get_region())

        async def _get_token() -> str:
            """Common token fetching logic."""
            return await client.get_token(
                provider_name=provider_name,
                agent_identity_token=await _get_workload_access_token(client),
                scopes=scopes,
                on_auth_url=on_auth_url,
                auth_flow=auth_flow,
                callback_url=_get_oauth2_callback_url(callback_url),
                force_authentication=force_authentication,
                token_poller=token_poller,
                custom_state=custom_state,
                custom_parameters=custom_parameters,
            )

        @wraps(func)
        async def async_wrapper(*args: Any, **kwargs_func: Any) -> Any:
            token = await _get_token()
            kwargs_func[into] = token
            return await func(*args, **kwargs_func)

        @wraps(func)
        def sync_wrapper(*args: Any, **kwargs_func: Any) -> Any:
            if _has_running_loop():
                # for async env, eg. runtime
                ctx = contextvars.copy_context()
                import concurrent.futures

                with concurrent.futures.ThreadPoolExecutor() as executor:
                    future = executor.submit(ctx.run, asyncio.run, _get_token())
                    token = future.result()
            else:
                # for sync env, eg. local dev
                token = asyncio.run(_get_token())

            kwargs_func[into] = token
            return func(*args, **kwargs_func)

        # Return appropriate wrapper based on function type
        if asyncio.iscoroutinefunction(func):
            return async_wrapper
        else:
            return sync_wrapper

    return decorator


def requires_iam_access_token(
    *,
    audience: List[str],
    signing_algorithm: str = "ES384",
    duration_seconds: int = 300,
    tags: Optional[List[Dict[str, str]]] = None,
    into: str = "access_token",
) -> Callable:
    """Decorator that fetches an AWS IAM JWT token before calling the decorated function.

    This decorator obtains a signed JWT from AWS STS using the GetWebIdentityToken API.
    The JWT can be used to authenticate with external services that support OIDC token
    validation. No client secrets are required - the token is signed by AWS.

    This is separate from @requires_access_token which uses AgentCore Identity for
    OAuth 2.0 flows. Use this decorator for M2M authentication with services that
    accept AWS-signed JWTs.

    Args:
        audience: List of intended token recipients (populates 'aud' claim in JWT).
                  Must match what the external service expects.
        signing_algorithm: Algorithm for signing the JWT.
                        'ES384' (default) or 'RS256'.
        duration_seconds: Token lifetime in seconds (60-3600, default 300).
        tags: Optional custom claims as [{'Key': str, 'Value': str}, ...].
              These are added to the JWT as additional claims.
        into: Parameter name to inject the token into (default: 'access_token').

    Returns:
        Decorator function that wraps the target function.

    Raises:
        ValueError: If parameters are invalid.
        RuntimeError: If AWS JWT federation is not enabled for the account.
        ClientError: If the STS API call fails.

    Example:
        @tool
        @requires_iam_access_token(
            audience=["https://api.example.com"],
            signing_algorithm="ES384",
            duration_seconds=300,
        )
        def call_external_api(query: str, *, access_token: str) -> str:
            '''Call external API with AWS JWT authentication.'''
            import requests
            response = requests.get(
                "https://api.example.com/data",
                headers={"Authorization": f"Bearer {access_token}"},
                params={"q": query},
            )
            return response.text

    Note:
        Before using this decorator, you must:
        1. Enable AWS IAM Outbound Web Identity Federation for your account
           (via `agentcore identity setup-aws-jwt` or IAM API)
        2. Ensure the execution role has `sts:GetWebIdentityToken` permission
        3. Configure the external service to trust your AWS account's issuer URL
    """
    # Validate parameters
    if not audience:
        raise ValueError("audience is required")
    if signing_algorithm not in ["ES384", "RS256"]:
        raise ValueError("signing_algorithm must be 'ES384' or 'RS256'")
    if not (60 <= duration_seconds <= 3600):
        raise ValueError("duration_seconds must be between 60 and 3600")

    logger = logging.getLogger(__name__)

    def _get_iam_jwt_token(region: str) -> str:
        """Get JWT from AWS STS - NO IdentityClient involved."""
        logger.info("Getting AWS IAM JWT token from STS...")
        sts_client = boto3.client("sts", region_name=region)

        params = {
            "Audience": audience,
            "SigningAlgorithm": signing_algorithm,
            "DurationSeconds": duration_seconds,
        }
        if tags:
            params["Tags"] = tags

        try:
            response = sts_client.get_web_identity_token(**params)
            logger.info("Successfully obtained AWS IAM JWT token")
            return response["WebIdentityToken"]
        except ClientError as e:
            error_code = e.response.get("Error", {}).get("Code", "")
            if error_code in ["FeatureDisabledException", "FeatureDisabled"]:
                raise RuntimeError("AWS IAM Outbound Web Identity Federation is not enabled.") from e
            logger.error("Failed to get AWS IAM JWT token: %s", str(e))
            raise

    def decorator(func: Callable) -> Callable:
        @wraps(func)
        async def async_wrapper(*args: Any, **kwargs_func: Any) -> Any:
            region = _get_region()
            token = _get_iam_jwt_token(region)
            kwargs_func[into] = token
            return await func(*args, **kwargs_func)

        @wraps(func)
        def sync_wrapper(*args: Any, **kwargs_func: Any) -> Any:
            region = _get_region()
            token = _get_iam_jwt_token(region)
            kwargs_func[into] = token
            return func(*args, **kwargs_func)

        if asyncio.iscoroutinefunction(func):
            return async_wrapper
        return sync_wrapper

    return decorator


def requires_api_key(*, provider_name: str, into: str = "api_key") -> Callable:
    """Decorator that fetches an API key before calling the decorated function.

    Args:
        provider_name: The credential provider name
        into: Parameter name to inject the API key into

    Returns:
        Decorator function
    """

    def decorator(func: Callable) -> Callable:
        client = IdentityClient(_get_region())

        async def _get_api_key():
            return await client.get_api_key(
                provider_name=provider_name,
                agent_identity_token=await _get_workload_access_token(client),
            )

        @wraps(func)
        async def async_wrapper(*args: Any, **kwargs: Any) -> Any:
            api_key = await _get_api_key()
            kwargs[into] = api_key
            return await func(*args, **kwargs)

        @wraps(func)
        def sync_wrapper(*args: Any, **kwargs: Any) -> Any:
            if _has_running_loop():
                # for async env, eg. runtime
                ctx = contextvars.copy_context()
                import concurrent.futures

                with concurrent.futures.ThreadPoolExecutor() as executor:
                    future = executor.submit(ctx.run, asyncio.run, _get_api_key())
                    api_key = future.result()
            else:
                # for sync env, eg. local dev
                api_key = asyncio.run(_get_api_key())

            kwargs[into] = api_key
            return func(*args, **kwargs)

        if asyncio.iscoroutinefunction(func):
            return async_wrapper
        else:
            return sync_wrapper

    return decorator


def _get_oauth2_callback_url(user_provided_oauth2_callback_url: Optional[str]):
    if user_provided_oauth2_callback_url:
        return user_provided_oauth2_callback_url

    return BedrockAgentCoreContext.get_oauth2_callback_url()


async def _get_workload_access_token(client: IdentityClient) -> str:
    token = BedrockAgentCoreContext.get_workload_access_token()
    if token is not None:
        return token
    else:
        # workload access token context var was not set, so we should be running in a local dev environment
        if os.getenv("DOCKER_CONTAINER") == "1":
            raise ValueError(
                "Workload access token has not been set. If invoking agent runtime via SIGV4 inbound auth, "
                "please specify the X-Amzn-Bedrock-AgentCore-Runtime-User-Id header and retry. "
                "For details, see - https://docs.aws.amazon.com/bedrock-agentcore/latest/devguide/runtime-oauth.html"
            )

        return await _set_up_local_auth(client)


async def _set_up_local_auth(client: IdentityClient) -> str:
    import json
    import uuid
    from pathlib import Path

    config_path = Path(".agentcore.json")
    workload_identity_name = None
    config = {}
    if config_path.exists():
        try:
            with open(config_path, "r", encoding="utf-8") as file:
                config = json.load(file) or {}
        except Exception:
            print("Could not find existing workload identity and user id")

    workload_identity_name = config.get("workload_identity_name")
    if workload_identity_name:
        print(f"Found existing workload identity from {config_path.absolute()}: {workload_identity_name}")
    else:
        workload_identity_name = client.create_workload_identity()["name"]
        print("Created a workload identity")

    user_id = config.get("user_id")
    if user_id:
        print(f"Found existing user id from {config_path.absolute()}: {user_id}")
    else:
        user_id = uuid.uuid4().hex[:8]
        print("Created an user id")

    try:
        config = {"workload_identity_name": workload_identity_name, "user_id": user_id}
        with open(config_path, "w", encoding="utf-8") as file:
            json.dump(config, file, indent=2)
    except Exception:
        print("Warning: could not write the created workload identity to file")

    return client.get_workload_access_token(workload_identity_name, user_id=user_id)["workloadAccessToken"]


def _get_region() -> str:
    region_env = os.getenv("AWS_REGION", None)
    if region_env is not None:
        return region_env

    return boto3.Session().region_name or "us-west-2"


def _has_running_loop() -> bool:
    try:
        asyncio.get_running_loop()
        return True
    except RuntimeError:
        return False



"""
Integration tests for AgentCore Memory Session Manager.

Run with: python -m pytest tests_integ/memory/integrations/test_session_manager.py -v
"""

import json
import logging
import os
import time
import uuid
from datetime import datetime, timezone

import pytest
from strands import Agent
from strands.types.session import Session, SessionAgent, SessionType

from bedrock_agentcore.memory import MemoryClient
from bedrock_agentcore.memory.integrations.strands.bedrock_converter import AgentCoreMemoryConverter
from bedrock_agentcore.memory.integrations.strands.config import AgentCoreMemoryConfig, RetrievalConfig
from bedrock_agentcore.memory.integrations.strands.session_manager import AgentCoreMemorySessionManager
from bedrock_agentcore.memory.models.filters import (
    EventMetadataFilter,
    LeftExpression,
    OperatorType,
    RightExpression,
)

logging.basicConfig(level=logging.INFO, format="%(asctime)s - %(levelname)s - %(message)s")
logger = logging.getLogger(__name__)

REGION = os.environ.get("BEDROCK_TEST_REGION", "us-east-1")


@pytest.mark.integration
class TestAgentCoreMemorySessionManager:
    """Integration tests for AgentCore Memory Session Manager."""

    @classmethod
    def setup_class(cls):
        """Set up test environment."""
        cls.region = os.environ.get("BEDROCK_TEST_REGION", "us-east-1")
        cls.client = MemoryClient(region_name=cls.region)

    @pytest.fixture(scope="session")
    def memory_client(self):
        """Create a memory client for testing."""
        return MemoryClient(region_name=REGION)

    @pytest.fixture(scope="session")
    def test_memory_stm(self, memory_client):
        """Create a test memory for integration tests."""
        memory_name = f"testmemorySTM{uuid.uuid4().hex[:8]}"
        memory = memory_client.create_memory_and_wait(
            name=memory_name, description="Test STM memory for integration tests", strategies=[]
        )
        yield memory
        # Cleanup
        try:
            memory_client.delete_memory(memory["id"])
        except Exception:
            pass  # Memory might already be deleted

    @pytest.fixture(scope="session")
    def test_memory_ltm(self, memory_client):
        """Create a test memory for integration tests."""
        memory_name = f"testmemoryLTM{uuid.uuid4().hex[:8]}"
        memory = memory_client.create_memory_and_wait(
            name=memory_name,
            description="Full-featured memory with all built-in strategies",
            strategies=[
                {
                    "summaryMemoryStrategy": {
                        "name": "SessionSummarizer",
                        "namespaces": ["/summaries/{actorId}/{sessionId}/"],
                    }
                },
                {
                    "userPreferenceMemoryStrategy": {
                        "name": "PreferenceLearner",
                        "namespaces": ["/preferences/{actorId}/"],
                    }
                },
                {"semanticMemoryStrategy": {"name": "FactExtractor", "namespaces": ["/facts/{actorId}/"]}},
            ],
        )
        yield memory
        try:
            memory_client.delete_memory(memory["id"])
        except Exception:
            pass  # Memory might already be deleted

    def test_session_manager_initialization(self, test_memory_stm):
        """Test session manager initialization."""
        session_config = AgentCoreMemoryConfig(
            memory_id=test_memory_stm["id"],
            session_id=f"test-session-{int(time.time())}",
            actor_id=f"test-actor-{int(time.time())}",
        )
        session_manager = AgentCoreMemorySessionManager(agentcore_memory_config=session_config, region_name=REGION)

        assert session_manager.config == session_config
        assert session_manager.memory_client is not None

    def test_agent_with_session_manager(self, test_memory_stm):
        """Test creating an agent with the session manager."""
        session_config = AgentCoreMemoryConfig(
            memory_id=test_memory_stm["id"],
            session_id=f"test-session-{int(time.time())}",
            actor_id=f"test-actor-{int(time.time())}",
        )
        session_manager = AgentCoreMemorySessionManager(agentcore_memory_config=session_config, region_name=REGION)

        agent = Agent(system_prompt="You are a helpful assistant.", session_manager=session_manager)

        assert agent._session_manager == session_manager

    @pytest.mark.flaky(reruns=2, reruns_delay=5)
    def test_conversation_persistence(self, test_memory_stm):
        """Test that conversations are persisted to memory."""
        session_config = AgentCoreMemoryConfig(
            memory_id=test_memory_stm["id"],
            session_id=f"test-session-{int(time.time())}",
            actor_id=f"test-actor-{int(time.time())}",
        )
        session_manager = AgentCoreMemorySessionManager(agentcore_memory_config=session_config, region_name=REGION)

        agent = Agent(system_prompt="You are a helpful assistant.", session_manager=session_manager)

        # Have a conversation
        response1 = agent("Hello, my name is John")
        assert response1 is not None

        time.sleep(2)  # throttling
        response2 = agent("What is my name?")
        assert response2 is not None
        assert "John" in response2.message["content"][0]["text"]

    @pytest.mark.flaky(reruns=2, reruns_delay=5)
    def test_session_manager_with_retrieval_config_adds_context(self, test_memory_ltm):
        """Test session manager with custom retrieval configuration."""
        actor_id = f"test-actor-{int(time.time())}"
        config = AgentCoreMemoryConfig(
            memory_id=test_memory_ltm["id"],
            session_id=f"test-session-{int(time.time())}",
            actor_id=actor_id,
            retrieval_config={"/preferences/{actorId}/": RetrievalConfig(top_k=5, relevance_score=0.7)},
        )

        session_manager = AgentCoreMemorySessionManager(agentcore_memory_config=config, region_name=REGION)

        agent = Agent(system_prompt="You are a helpful assistant.", session_manager=session_manager)

        response1 = agent("I like sushi with tuna")
        assert response1 is not None
        self.client.wait_for_memories(
            memory_id=test_memory_ltm["id"],
            namespace=f"/preferences/{actor_id}/",
            test_query="sushi",
            max_wait=180,
            poll_interval=10,
        )

        response2 = agent("What do I like to eat?")
        assert response2 is not None
        assert "sushi" in str(agent.messages)
        assert "<user_context>" in str(agent.messages)

    @pytest.mark.flaky(reruns=2, reruns_delay=5)
    def test_multiple_namespace_retrieval_config(self, test_memory_ltm):
        """Test session manager with multiple namespace retrieval configurations."""
        actor_id = f"test-actor-{int(time.time())}"
        config = AgentCoreMemoryConfig(
            memory_id=test_memory_ltm["id"],
            session_id=f"test-session-{int(time.time())}",
            actor_id=actor_id,
            retrieval_config={
                "/preferences/{actorId}/": RetrievalConfig(top_k=5, relevance_score=0.7),
                "/facts/{actorId}/": RetrievalConfig(top_k=10, relevance_score=0.3),
                "/summaries/{actorId}/{sessionId}/": RetrievalConfig(top_k=5, relevance_score=0.5),
            },
        )

        session_manager = AgentCoreMemorySessionManager(agentcore_memory_config=config, region_name=REGION)

        assert len(session_manager.config.retrieval_config) == 3
        agent = Agent(
            system_prompt="You are a helpful assistant that understands user preferences.",
            session_manager=session_manager,
        )

        response1 = agent("I like sushi with tuna")
        assert response1 is not None
        self.client.wait_for_memories(
            memory_id=test_memory_ltm["id"],
            namespace=f"/preferences/{actor_id}/",
            test_query="sushi",
            max_wait=180,
            poll_interval=10,
        )

        response2 = agent("What do I like to eat?")
        assert response2 is not None
        assert "sushi" in str(agent.messages)
        assert "<user_context>" in str(agent.messages)

    def test_session_manager_error_handling(self):
        """Test session manager error handling with invalid configuration."""
        with pytest.raises(Exception):  # noqa: B017
            # Invalid memory ID should raise an error
            config = AgentCoreMemoryConfig(
                memory_id="invalid-memory-id", session_id="test-session", actor_id="test-actor"
            )

            session_manager = AgentCoreMemorySessionManager(agentcore_memory_config=config, region_name=REGION)

            # This should fail when trying to use the session manager
            agent = Agent(system_prompt="Test", session_manager=session_manager)
            agent("Test message")

    def test_legacy_event_migration(self, test_memory_stm, memory_client):
        """Test that legacy events with prefixed actorIds are migrated to metadata format.

        The constructor calls read_session which creates a metadata-path session if none exists.
        To test legacy migration, we create the legacy event BEFORE constructing the session manager,
        so the constructor's read_session finds it via the fallback and migrates it on first access.
        """
        session_id = f"test-legacy-{uuid.uuid4().hex[:8]}"
        actor_id = f"test-actor-{uuid.uuid4().hex[:8]}"

        # --- Session migration ---
        # Create a legacy session event BEFORE constructing the session manager.
        # Legacy events use blob payloads with the session data, so we use gmdp_client directly.
        legacy_session_actor_id = f"session_{session_id}"
        session_data = Session(session_id=session_id, session_type=SessionType.AGENT)
        memory_client.gmdp_client.create_event(
            memoryId=test_memory_stm["id"],
            actorId=legacy_session_actor_id,
            sessionId=session_id,
            payload=[{"blob": json.dumps(session_data.to_dict())}],
            eventTimestamp=datetime.now(timezone.utc),
        )

        # Verify legacy event exists before migration
        legacy_events_before = memory_client.list_events(
            memory_id=test_memory_stm["id"],
            actor_id=legacy_session_actor_id,
            session_id=session_id,
        )
        assert len(legacy_events_before) >= 1

        # Constructing the session manager triggers read_session in __init__,
        # which should find the legacy event, migrate it, and delete the old one
        config = AgentCoreMemoryConfig(
            memory_id=test_memory_stm["id"],
            session_id=session_id,
            actor_id=actor_id,
        )
        session_manager = AgentCoreMemorySessionManager(agentcore_memory_config=config, region_name=REGION)

        # Verify migration: legacy event should be deleted
        legacy_events_after = memory_client.list_events(
            memory_id=test_memory_stm["id"],
            actor_id=legacy_session_actor_id,
            session_id=session_id,
        )
        assert len(legacy_events_after) == 0

        # Verify migration: read_session finds it via the new metadata path
        read_session_result = session_manager.read_session(session_id)
        assert read_session_result is not None
        assert read_session_result.session_id == session_id

        # --- Agent migration ---
        agent_id = f"test-agent-{uuid.uuid4().hex[:8]}"
        legacy_agent_actor_id = f"agent_{agent_id}"
        agent_data = SessionAgent(
            agent_id=agent_id,
            state={"key": "value"},
            conversation_manager_state={},
        )
        memory_client.gmdp_client.create_event(
            memoryId=test_memory_stm["id"],
            actorId=legacy_agent_actor_id,
            sessionId=session_id,
            payload=[{"blob": json.dumps(agent_data.to_dict())}],
            eventTimestamp=datetime.now(timezone.utc),
        )

        # read_agent should find via fallback and migrate
        read_agent_result = session_manager.read_agent(session_id, agent_id)
        assert read_agent_result is not None
        assert read_agent_result.agent_id == agent_id

        # Verify migration: legacy event should be deleted
        legacy_agent_events = memory_client.list_events(
            memory_id=test_memory_stm["id"],
            actor_id=legacy_agent_actor_id,
            session_id=session_id,
        )
        assert len(legacy_agent_events) == 0

    # endregion Event metadata integration tests

    # region End-to-end agent with batching tests

    @pytest.mark.flaky(reruns=2, reruns_delay=5)
    def test_agent_conversation_with_context_manager(self, test_memory_stm):
        """Test that Agent messages are flushed when the context manager exits, and session resume loads them."""
        session_id = f"test-agent-ctx-{uuid.uuid4().hex[:8]}"
        actor_id = f"test-actor-{uuid.uuid4().hex[:8]}"

        config = AgentCoreMemoryConfig(
            memory_id=test_memory_stm["id"],
            session_id=session_id,
            actor_id=actor_id,
            batch_size=10,
        )

        # Use context manager — __exit__ calls _flush_messages() which is blocking
        with AgentCoreMemorySessionManager(agentcore_memory_config=config, region_name=REGION) as sm:
            agent = Agent(system_prompt="You are a helpful assistant.", session_manager=sm)
            response1 = agent("Hello, my name is Bob")
            assert response1 is not None

        # After __exit__, buffered messages have been flushed (blocking).
        # Resume session with a new session manager to verify persistence.
        config2 = AgentCoreMemoryConfig(
            memory_id=test_memory_stm["id"],
            session_id=session_id,
            actor_id=actor_id,
            batch_size=10,
        )
        sm2 = AgentCoreMemorySessionManager(agentcore_memory_config=config2, region_name=REGION)
        agent2 = Agent(system_prompt="You are a helpful assistant.", session_manager=sm2)

        response2 = agent2("What is my name?")
        assert response2 is not None
        assert "Bob" in response2.message["content"][0]["text"]

        sm2.close()

    @pytest.mark.flaky(reruns=2, reruns_delay=5)
    def test_agent_multi_turn_with_batching(self, test_memory_stm):
        """Test that a multi-turn conversation within a single Agent works with batching."""
        session_id = f"test-agent-multi-{uuid.uuid4().hex[:8]}"
        actor_id = f"test-actor-{uuid.uuid4().hex[:8]}"

        config = AgentCoreMemoryConfig(
            memory_id=test_memory_stm["id"],
            session_id=session_id,
            actor_id=actor_id,
            batch_size=10,
        )
        session_manager = AgentCoreMemorySessionManager(agentcore_memory_config=config, region_name=REGION)

        agent = Agent(system_prompt="You are a helpful assistant.", session_manager=session_manager)

        agent("Hello, my name is Charlie")
        agent("I live in Seattle")
        response3 = agent("What is my name and where do I live?")
        assert response3 is not None
        response_text = response3.message["content"][0]["text"]
        assert "Charlie" in response_text
        assert "Seattle" in response_text

        # Flush remaining buffered messages (blocking)
        session_manager.close()

        # Verify batched messages are persisted — filter out state events
        message_filter = EventMetadataFilter.build_expression(
            left_operand=LeftExpression.build("stateType"),
            operator=OperatorType.NOT_EXISTS,
        )
        events = session_manager.memory_client.list_events(
            memory_id=test_memory_stm["id"],
            actor_id=actor_id,
            session_id=session_id,
            event_metadata=[message_filter],
        )

        # Convert events back to messages and verify all turns are present
        messages = AgentCoreMemoryConverter.events_to_messages(events)
        # At least 3 user + 3 assistant messages
        assert len(messages) >= 6

    # endregion End-to-end agent with batching tests

    # region Event metadata integration tests

    def test_create_message_with_metadata_and_filter(self, test_memory_stm, memory_client):
        """Test that user-supplied metadata is persisted and that filters actually exclude non-matching events."""
        session_id = f"test-meta-{uuid.uuid4().hex[:8]}"
        actor_id = f"test-actor-{uuid.uuid4().hex[:8]}"

        config = AgentCoreMemoryConfig(
            memory_id=test_memory_stm["id"],
            session_id=session_id,
            actor_id=actor_id,
            default_metadata={"project": {"stringValue": "atlas"}, "env": {"stringValue": "test"}},
        )
        session_manager = AgentCoreMemorySessionManager(agentcore_memory_config=config, region_name=REGION)

        agent = Agent(system_prompt="You are a helpful assistant.", session_manager=session_manager)
        agent("Hello, remember my project is Atlas")

        # Get ALL events (unfiltered) to know the total count
        all_events = memory_client.list_events(
            memory_id=test_memory_stm["id"],
            actor_id=actor_id,
            session_id=session_id,
        )
        assert len(all_events) >= 1  # Sanity: events exist

        # Positive filter: query events matching our custom metadata
        project_filter = EventMetadataFilter.build_expression(
            left_operand=LeftExpression.build("project"),
            operator=OperatorType.EQUALS_TO,
            right_operand=RightExpression.build("atlas"),
        )
        matching_events = memory_client.list_events(
            memory_id=test_memory_stm["id"],
            actor_id=actor_id,
            session_id=session_id,
            event_metadata=[project_filter],
        )
        assert len(matching_events) >= 1

        # Verify metadata values round-trip correctly
        for event in matching_events:
            meta = event.get("metadata", {})
            assert meta.get("project", {}).get("stringValue") == "atlas"
            assert meta.get("env", {}).get("stringValue") == "test"

        # Negative filter: query with a value that was never written
        wrong_filter = EventMetadataFilter.build_expression(
            left_operand=LeftExpression.build("project"),
            operator=OperatorType.EQUALS_TO,
            right_operand=RightExpression.build("nonexistent_project"),
        )
        non_matching_events = memory_client.list_events(
            memory_id=test_memory_stm["id"],
            actor_id=actor_id,
            session_id=session_id,
            event_metadata=[wrong_filter],
        )
        assert len(non_matching_events) == 0, (
            f"Expected 0 events for nonexistent metadata value, got {len(non_matching_events)}"
        )

        # The positive set should be a strict subset of all events
        # (all_events includes state events that don't have user metadata)
        assert len(matching_events) < len(all_events)

    def test_metadata_survives_session_resume(self, test_memory_stm, memory_client):
        """Events with metadata written by one session manager are filterable from another."""
        session_id = f"test-meta-resume-{uuid.uuid4().hex[:8]}"
        actor_id = f"test-actor-{uuid.uuid4().hex[:8]}"

        # First session: write messages with metadata
        config1 = AgentCoreMemoryConfig(
            memory_id=test_memory_stm["id"],
            session_id=session_id,
            actor_id=actor_id,
            default_metadata={"source": {"stringValue": "session1"}},
        )
        sm1 = AgentCoreMemorySessionManager(agentcore_memory_config=config1, region_name=REGION)
        agent1 = Agent(system_prompt="You are a helpful assistant.", session_manager=sm1)
        agent1("My favourite colour is blue")

        # Second session: resume without metadata configured
        config2 = AgentCoreMemoryConfig(
            memory_id=test_memory_stm["id"],
            session_id=session_id,
            actor_id=actor_id,
        )
        sm2 = AgentCoreMemorySessionManager(agentcore_memory_config=config2, region_name=REGION)

        # Positive: filter matches metadata written by session 1
        source_filter = EventMetadataFilter.build_expression(
            left_operand=LeftExpression.build("source"),
            operator=OperatorType.EQUALS_TO,
            right_operand=RightExpression.build("session1"),
        )
        matching = sm2.memory_client.list_events(
            memory_id=test_memory_stm["id"],
            actor_id=actor_id,
            session_id=session_id,
            event_metadata=[source_filter],
        )
        assert len(matching) >= 1

        # Negative: filter for a different source value returns nothing
        wrong_source_filter = EventMetadataFilter.build_expression(
            left_operand=LeftExpression.build("source"),
            operator=OperatorType.EQUALS_TO,
            right_operand=RightExpression.build("session999"),
        )
        not_matching = sm2.memory_client.list_events(
            memory_id=test_memory_stm["id"],
            actor_id=actor_id,
            session_id=session_id,
            event_metadata=[wrong_source_filter],
        )
        assert len(not_matching) == 0

    def test_metadata_provider_attaches_dynamic_trace_id(self, test_memory_stm, memory_client):
        """metadata_provider injects a different traceId per invocation and events are filterable by each."""
        session_id = f"test-meta-prov-{uuid.uuid4().hex[:8]}"
        actor_id = f"test-actor-{uuid.uuid4().hex[:8]}"

        current_trace = {"traceId": {"stringValue": "trace-AAA"}}

        config = AgentCoreMemoryConfig(
            memory_id=test_memory_stm["id"],
            session_id=session_id,
            actor_id=actor_id,
            metadata_provider=lambda: dict(current_trace),
        )
        sm = AgentCoreMemorySessionManager(agentcore_memory_config=config, region_name=REGION)
        agent = Agent(system_prompt="You are a helpful assistant.", session_manager=sm)

        # First invocation with trace-AAA
        agent("Hello from trace AAA")

        # Switch trace for second invocation
        current_trace["traceId"] = {"stringValue": "trace-BBB"}
        agent("Hello from trace BBB")

        # Filter for trace-AAA — should find events
        filter_aaa = EventMetadataFilter.build_expression(
            left_operand=LeftExpression.build("traceId"),
            operator=OperatorType.EQUALS_TO,
            right_operand=RightExpression.build("trace-AAA"),
        )
        events_aaa = memory_client.list_events(
            memory_id=test_memory_stm["id"],
            actor_id=actor_id,
            session_id=session_id,
            event_metadata=[filter_aaa],
        )
        assert len(events_aaa) >= 1
        for e in events_aaa:
            assert e.get("metadata", {}).get("traceId", {}).get("stringValue") == "trace-AAA"

        # Filter for trace-BBB — should find events
        filter_bbb = EventMetadataFilter.build_expression(
            left_operand=LeftExpression.build("traceId"),
            operator=OperatorType.EQUALS_TO,
            right_operand=RightExpression.build("trace-BBB"),
        )
        events_bbb = memory_client.list_events(
            memory_id=test_memory_stm["id"],
            actor_id=actor_id,
            session_id=session_id,
            event_metadata=[filter_bbb],
        )
        assert len(events_bbb) >= 1
        for e in events_bbb:
            assert e.get("metadata", {}).get("traceId", {}).get("stringValue") == "trace-BBB"

        # Negative: nonexistent trace returns nothing
        filter_none = EventMetadataFilter.build_expression(
            left_operand=LeftExpression.build("traceId"),
            operator=OperatorType.EQUALS_TO,
            right_operand=RightExpression.build("trace-DOES-NOT-EXIST"),
        )
        events_none = memory_client.list_events(
            memory_id=test_memory_stm["id"],
            actor_id=actor_id,
            session_id=session_id,
            event_metadata=[filter_none],
        )
        assert len(events_none) == 0

        # The two traces should be disjoint sets
        aaa_ids = {e["eventId"] for e in events_aaa}
        bbb_ids = {e["eventId"] for e in events_bbb}
        assert aaa_ids.isdisjoint(bbb_ids), "trace-AAA and trace-BBB events should not overlap"

    # endregion Event metadata integration tests



"""BedrockAgentCore Runtime Package.

This package contains the core runtime components for Bedrock AgentCore applications:
- BedrockAgentCoreApp: Main application class
- RequestContext: HTTP request context
- BedrockAgentCoreContext: Agent identity context
"""

from .agent_core_runtime_client import AgentCoreRuntimeClient
from .app import BedrockAgentCoreApp
from .context import BedrockAgentCoreContext, RequestContext
from .models import PingStatus

__all__ = [
    "AgentCoreRuntimeClient",
    "AGUIApp",
    "BedrockAgentCoreApp",
    "BedrockCallContextBuilder",
    "RequestContext",
    "BedrockAgentCoreContext",
    "PingStatus",
    "build_a2a_app",
    "build_ag_ui_app",
    "build_runtime_url",
    "serve_a2a",
    "serve_ag_ui",
]


def __getattr__(name: str):
    """Lazy imports for A2A and AG-UI symbols so optional dependencies are not required at import time."""
    _a2a_exports = {"BedrockCallContextBuilder", "build_a2a_app", "build_runtime_url", "serve_a2a"}
    if name in _a2a_exports:
        from . import a2a as _a2a_module

        return getattr(_a2a_module, name)

    _ag_ui_exports = {"AGUIApp", "build_ag_ui_app", "serve_ag_ui"}
    if name in _ag_ui_exports:
        from . import ag_ui as _ag_ui_module

        return getattr(_ag_ui_module, name)

    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")



import asyncio
import json
import logging
import textwrap

import websockets

from tests_integ.runtime.base_test import BaseSDKRuntimeTest, start_agent_server

logger = logging.getLogger("sdk-runtime-websocket-test")

WEBSOCKET_PORT = 8081


class TestSDKWebSocketAgent(BaseSDKRuntimeTest):
    def setup(self):
        self.agent_module = "websocket_agent"
        with open(self.agent_module + ".py", "w") as file:
            content = textwrap.dedent("""
                from bedrock_agentcore import BedrockAgentCoreApp

                app = BedrockAgentCoreApp(debug=True)

                @app.websocket
                async def websocket_handler(websocket, context):
                    await websocket.accept()

                    # Echo server - receive and respond to messages
                    try:
                        while True:
                            data = await websocket.receive_json()

                            # Handle different message types
                            if data.get("action") == "echo":
                                await websocket.send_json({
                                    "type": "echo_response",
                                    "message": data.get("message"),
                                    "session_id": context.session_id
                                })
                            elif data.get("action") == "stream":
                                # Stream multiple messages
                                count = data.get("count", 3)
                                for i in range(count):
                                    await websocket.send_json({
                                        "type": "stream_chunk",
                                        "chunk_id": i,
                                        "data": f"Chunk {i+1} of {count}"
                                    })
                                await websocket.send_json({"type": "stream_complete"})
                            elif data.get("action") == "close":
                                await websocket.send_json({"type": "closing"})
                                break
                    except Exception as e:
                        await websocket.send_json({"type": "error", "message": str(e)})
                    finally:
                        await websocket.close()

                app.run(port=8081)
            """).strip()
            file.write(content)

    def run_test(self):
        with start_agent_server(self.agent_module, port=WEBSOCKET_PORT):
            ws_endpoint = f"ws://127.0.0.1:{WEBSOCKET_PORT}/ws"

            # Run async WebSocket tests
            asyncio.run(self._test_websocket_echo(ws_endpoint))
            asyncio.run(self._test_websocket_streaming(ws_endpoint))
            asyncio.run(self._test_websocket_with_session(ws_endpoint))

    async def _test_websocket_echo(self, ws_endpoint):
        """Test basic WebSocket echo functionality."""
        logger.info("Testing WebSocket echo...")

        async with websockets.connect(ws_endpoint) as websocket:
            # Send echo request
            await websocket.send(json.dumps({"action": "echo", "message": "Hello WebSocket!"}))

            # Receive echo response
            response = await websocket.recv()
            data = json.loads(response)

            logger.info("Echo response: %s", data)
            assert data["type"] == "echo_response"
            assert data["message"] == "Hello WebSocket!"

            # Close connection
            await websocket.send(json.dumps({"action": "close"}))
            closing_msg = await websocket.recv()
            assert json.loads(closing_msg)["type"] == "closing"

    async def _test_websocket_streaming(self, ws_endpoint):
        """Test WebSocket streaming functionality."""
        logger.info("Testing WebSocket streaming...")

        async with websockets.connect(ws_endpoint) as websocket:
            # Request stream of 5 messages
            await websocket.send(json.dumps({"action": "stream", "count": 5}))

            # Receive streamed chunks
            chunks = []
            for _ in range(5):
                response = await websocket.recv()
                chunk = json.loads(response)
                logger.info("Received chunk: %s", chunk)
                assert chunk["type"] == "stream_chunk"
                chunks.append(chunk)

            # Receive completion message
            complete_msg = await websocket.recv()
            completion = json.loads(complete_msg)
            assert completion["type"] == "stream_complete"

            # Verify all chunks received
            assert len(chunks) == 5
            assert chunks[0]["chunk_id"] == 0
            assert chunks[4]["chunk_id"] == 4

            # Close connection
            await websocket.send(json.dumps({"action": "close"}))

    async def _test_websocket_with_session(self, ws_endpoint):
        """Test WebSocket with session ID in headers."""
        logger.info("Testing WebSocket with session ID...")

        # Add session ID header
        extra_headers = [("X-Amzn-Bedrock-AgentCore-Runtime-Session-Id", "test-session-123")]

        async with websockets.connect(ws_endpoint, additional_headers=extra_headers) as websocket:
            # Send echo request
            await websocket.send(json.dumps({"action": "echo", "message": "Session test"}))

            # Receive response with session ID
            response = await websocket.recv()
            data = json.loads(response)

            logger.info("Response with session: %s", data)
            assert data["session_id"] == "test-session-123"

            # Close connection
            await websocket.send(json.dumps({"action": "close"}))


def test(tmp_path):
    TestSDKWebSocketAgent().run(tmp_path)



import contextvars
import uuid
from unittest.mock import patch

from a2a.server.agent_execution import AgentExecutor, RequestContext
from a2a.server.events import EventQueue
from a2a.server.tasks import InMemoryTaskStore, TaskUpdater
from a2a.types import AgentCapabilities, AgentCard, AgentSkill, Part, TextPart
from a2a.utils import new_task
from starlette.testclient import TestClient

from bedrock_agentcore.runtime.a2a import (
    BedrockCallContextBuilder,
    build_a2a_app,
    build_runtime_url,
    serve_a2a,
)
from bedrock_agentcore.runtime.context import BedrockAgentCoreContext
from bedrock_agentcore.runtime.models import PingStatus


class _EchoExecutor(AgentExecutor):
    """Echoes user input back as an artifact. Records call_context for inspection."""

    def __init__(self):
        self.last_call_context = None

    async def execute(self, context: RequestContext, event_queue: EventQueue) -> None:
        self.last_call_context = context.call_context
        task = context.current_task or new_task(context.message)
        if not context.current_task:
            await event_queue.enqueue_event(task)
        updater = TaskUpdater(event_queue, task.id, task.context_id)
        user_text = context.get_user_input()
        await updater.add_artifact([Part(root=TextPart(text=f"echo: {user_text}"))])
        await updater.complete()

    async def cancel(self, context: RequestContext, event_queue: EventQueue) -> None:
        pass


def _make_agent_card() -> AgentCard:
    return AgentCard(
        name="test-agent",
        description="A test agent",
        url="http://localhost:9000",
        version="1.0.0",
        capabilities=AgentCapabilities(streaming=True),
        skills=[AgentSkill(id="echo", name="echo", description="Echoes input", tags=["echo"])],
        default_input_modes=["text"],
        default_output_modes=["text"],
    )


def _jsonrpc_request(method: str, params: dict | None = None) -> dict:
    body: dict = {"jsonrpc": "2.0", "method": method, "id": 1}
    if params is not None:
        body["params"] = params
    return body


def _send_message_params(text: str = "hello") -> dict:
    return {
        "message": {
            "message_id": str(uuid.uuid4()),
            "role": "user",
            "parts": [{"kind": "text", "text": text}],
        }
    }


class TestBuildA2AApp:
    def test_ping_returns_healthy_with_timestamp(self):
        app = build_a2a_app(_EchoExecutor(), _make_agent_card())
        client = TestClient(app)
        resp = client.get("/ping")
        assert resp.status_code == 200
        body = resp.json()
        assert body["status"] == "Healthy"
        assert isinstance(body["time_of_last_update"], int)
        assert body["time_of_last_update"] > 0

    def test_custom_ping_handler(self):
        app = build_a2a_app(
            _EchoExecutor(),
            _make_agent_card(),
            ping_handler=lambda: PingStatus.HEALTHY_BUSY,
        )
        client = TestClient(app)
        resp = client.get("/ping")
        assert resp.status_code == 200
        assert resp.json()["status"] == "HealthyBusy"

    def test_ping_fallback_on_error(self):
        def bad_ping():
            raise RuntimeError("boom")

        app = build_a2a_app(
            _EchoExecutor(),
            _make_agent_card(),
            ping_handler=bad_ping,
        )
        client = TestClient(app)
        resp = client.get("/ping")
        assert resp.status_code == 200
        assert resp.json()["status"] == "Healthy"

    def test_agent_card_returns_card_data(self):
        app = build_a2a_app(_EchoExecutor(), _make_agent_card())
        client = TestClient(app)
        resp = client.get("/.well-known/agent-card.json")
        assert resp.status_code == 200
        body = resp.json()
        assert body["name"] == "test-agent"
        assert body["version"] == "1.0.0"
        assert body["skills"][0]["id"] == "echo"

    def test_message_send_executes_and_returns_completed_task(self):
        """Verify the full RPC path: JSON-RPC request -> executor -> completed task."""
        app = build_a2a_app(_EchoExecutor(), _make_agent_card())
        client = TestClient(app, raise_server_exceptions=False)
        resp = client.post("/", json=_jsonrpc_request("message/send", _send_message_params("unit-test")))
        assert resp.status_code == 200
        body = resp.json()
        assert "result" in body
        task = body["result"]
        assert task["status"]["state"] == "completed"
        assert task["artifacts"][0]["parts"][0]["text"] == "echo: unit-test"

    def test_custom_task_store_is_used_for_persistence(self):
        """Verify that a custom task_store actually stores the task."""
        store = InMemoryTaskStore()
        app = build_a2a_app(_EchoExecutor(), _make_agent_card(), task_store=store)
        client = TestClient(app, raise_server_exceptions=False)

        # Send a message so a task gets created
        resp = client.post("/", json=_jsonrpc_request("message/send", _send_message_params("store-test")))
        task_id = resp.json()["result"]["id"]

        # Retrieve from the same store via tasks/get
        resp2 = client.post("/", json=_jsonrpc_request("tasks/get", {"id": task_id}))
        assert resp2.json()["result"]["id"] == task_id

    def test_agent_card_url_auto_populated_from_env(self):
        """When AGENTCORE_RUNTIME_URL is set, agent_card.url is overridden."""
        card = _make_agent_card()
        assert card.url == "http://localhost:9000"
        with patch.dict("os.environ", {"AGENTCORE_RUNTIME_URL": "https://deployed.example.com/"}):
            build_a2a_app(_EchoExecutor(), card)
        assert card.url == "https://deployed.example.com/"

    def test_agent_card_url_unchanged_without_env(self):
        """When AGENTCORE_RUNTIME_URL is not set, agent_card.url stays as-is."""
        card = _make_agent_card()
        original_url = card.url
        with patch.dict("os.environ", {}, clear=False):
            import os

            os.environ.pop("AGENTCORE_RUNTIME_URL", None)
            build_a2a_app(_EchoExecutor(), card)
        assert card.url == original_url

    def test_auto_builds_card_when_none_provided(self):
        """When agent_card is omitted, a default card is built automatically."""
        app = build_a2a_app(_EchoExecutor())
        client = TestClient(app)
        resp = client.get("/.well-known/agent-card.json")
        assert resp.status_code == 200
        body = resp.json()
        assert body["name"] == "agent"
        assert body["url"] == "http://localhost:9000/"

    def test_auto_builds_card_from_strands_executor(self):
        """When executor has .agent with name/description, card is built from it."""

        class _FakeAgent:
            name = "My Strands Agent"
            description = "Does cool stuff"

        executor = _EchoExecutor()
        executor.agent = _FakeAgent()  # type: ignore[attr-defined]

        app = build_a2a_app(executor)
        client = TestClient(app)
        resp = client.get("/.well-known/agent-card.json")
        body = resp.json()
        assert body["name"] == "My Strands Agent"
        assert body["description"] == "Does cool stuff"

    def test_auto_card_uses_runtime_url_env(self):
        """Auto-built card picks up AGENTCORE_RUNTIME_URL."""
        with patch.dict("os.environ", {"AGENTCORE_RUNTIME_URL": "https://prod.example.com/"}):
            app = build_a2a_app(_EchoExecutor())
        client = TestClient(app)
        resp = client.get("/.well-known/agent-card.json")
        assert resp.json()["url"] == "https://prod.example.com/"


class TestBedrockCallContextBuilder:
    def _run_in_isolated_context(self, fn):
        """Run fn in a fresh contextvars.Context so tests don't leak state."""
        ctx = contextvars.copy_context()
        return ctx.run(fn)

    def test_extracts_all_bedrock_headers(self):
        """Verify every Bedrock header is extracted into both ServerCallContext.state and contextvars."""

        def _test():
            from starlette.requests import Request

            scope = {
                "type": "http",
                "method": "POST",
                "path": "/",
                "headers": [
                    (b"x-amzn-bedrock-agentcore-runtime-session-id", b"sess-abc"),
                    (b"x-amzn-bedrock-agentcore-runtime-request-id", b"req-456"),
                    (b"workloadaccesstoken", b"tok-xyz"),
                    (b"oauth2callbackurl", b"https://callback.example.com"),
                    (b"x-amzn-bedrock-agentcore-runtime-custom-foo", b"bar"),
                    (b"authorization", b"Bearer test-token"),
                ],
                "query_string": b"",
            }
            request = Request(scope)
            builder = BedrockCallContextBuilder()
            result = builder.build(request)

            # Check ServerCallContext state dict
            assert result.state["session_id"] == "sess-abc"
            assert result.state["request_id"] == "req-456"
            assert result.state["workload_access_token"] == "tok-xyz"
            assert result.state["oauth2_callback_url"] == "https://callback.example.com"

            # Check contextvars were actually set
            assert BedrockAgentCoreContext.get_session_id() == "sess-abc"
            assert BedrockAgentCoreContext.get_request_id() == "req-456"
            assert BedrockAgentCoreContext.get_workload_access_token() == "tok-xyz"
            assert BedrockAgentCoreContext.get_oauth2_callback_url() == "https://callback.example.com"

            # Check custom + authorization headers collected
            headers = BedrockAgentCoreContext.get_request_headers()
            assert headers is not None
            assert "Authorization" in headers or "authorization" in headers
            custom_found = any(k.lower().startswith("x-amzn-bedrock-agentcore-runtime-custom-") for k in headers)
            assert custom_found

        self._run_in_isolated_context(_test)

    def test_auto_generates_request_id_when_missing(self):
        def _test():
            from starlette.requests import Request

            scope = {
                "type": "http",
                "method": "POST",
                "path": "/",
                "headers": [],
                "query_string": b"",
            }
            request = Request(scope)
            builder = BedrockCallContextBuilder()
            result = builder.build(request)

            generated_id = result.state["request_id"]
            assert generated_id is not None
            # Verify it's a valid UUID
            uuid.UUID(generated_id)
            assert BedrockAgentCoreContext.get_request_id() == generated_id

        self._run_in_isolated_context(_test)

    def test_optional_headers_omitted_from_state_when_absent(self):
        """When optional headers (token, oauth2) are missing, they must not appear in state."""

        def _test():
            from starlette.requests import Request

            scope = {
                "type": "http",
                "method": "POST",
                "path": "/",
                "headers": [
                    (b"x-amzn-bedrock-agentcore-runtime-session-id", b"sess-only"),
                ],
                "query_string": b"",
            }
            request = Request(scope)
            builder = BedrockCallContextBuilder()
            result = builder.build(request)

            assert result.state["session_id"] == "sess-only"
            assert "workload_access_token" not in result.state
            assert "oauth2_callback_url" not in result.state

        self._run_in_isolated_context(_test)

    def test_headers_reach_executor_through_full_app(self):
        """End-to-end: HTTP headers -> CallContextBuilder -> a2a-sdk -> executor.call_context."""

        def _test():
            executor = _EchoExecutor()
            builder = BedrockCallContextBuilder()
            app = build_a2a_app(executor, _make_agent_card(), context_builder=builder)
            client = TestClient(app, raise_server_exceptions=False)

            client.post(
                "/",
                json=_jsonrpc_request("message/send", _send_message_params("ctx-test")),
                headers={
                    "X-Amzn-Bedrock-AgentCore-Runtime-Session-Id": "unit-sess",
                    "X-Amzn-Bedrock-AgentCore-Runtime-Request-Id": "unit-req",
                    "WorkloadAccessToken": "unit-token",
                },
            )

            ctx = executor.last_call_context
            assert ctx is not None
            assert ctx.state["session_id"] == "unit-sess"
            assert ctx.state["request_id"] == "unit-req"
            assert ctx.state["workload_access_token"] == "unit-token"

        self._run_in_isolated_context(_test)


class TestServeA2A:
    @patch("uvicorn.run")
    def test_default_localhost(self, mock_uvicorn_run):
        with patch.dict("os.environ", {}, clear=False):
            with patch("os.path.exists", return_value=False):
                serve_a2a(_EchoExecutor(), _make_agent_card())
        kw = mock_uvicorn_run.call_args[1]
        assert kw["host"] == "127.0.0.1"
        assert kw["port"] == 9000

    @patch("uvicorn.run")
    def test_docker_detection_dockerenv(self, mock_uvicorn_run):
        with patch("os.path.exists", return_value=True):
            serve_a2a(_EchoExecutor(), _make_agent_card())
        assert mock_uvicorn_run.call_args[1]["host"] == "0.0.0.0"

    @patch("uvicorn.run")
    def test_docker_detection_env_var(self, mock_uvicorn_run):
        with patch.dict("os.environ", {"DOCKER_CONTAINER": "true"}):
            with patch("os.path.exists", return_value=False):
                serve_a2a(_EchoExecutor(), _make_agent_card())
        assert mock_uvicorn_run.call_args[1]["host"] == "0.0.0.0"

    @patch("uvicorn.run")
    def test_custom_host_port(self, mock_uvicorn_run):
        serve_a2a(_EchoExecutor(), _make_agent_card(), host="10.0.0.1", port=8888)
        kw = mock_uvicorn_run.call_args[1]
        assert kw["host"] == "10.0.0.1"
        assert kw["port"] == 8888

    @patch("uvicorn.run")
    def test_kwargs_passthrough(self, mock_uvicorn_run):
        serve_a2a(
            _EchoExecutor(),
            _make_agent_card(),
            host="127.0.0.1",
            workers=4,
            log_level="debug",
        )
        kw = mock_uvicorn_run.call_args[1]
        assert kw["workers"] == 4
        assert kw["log_level"] == "debug"

    @patch("uvicorn.run")
    def test_serve_without_agent_card(self, mock_uvicorn_run):
        """serve_a2a works with just an executor, no card needed."""
        with patch("os.path.exists", return_value=False):
            serve_a2a(_EchoExecutor())
        mock_uvicorn_run.assert_called_once()


class TestBuildRuntimeUrl:
    def test_builds_url_from_arn(self):
        arn = "arn:aws:bedrock-agentcore:us-east-1:123456789012:runtime/my-agent-abc123"
        url = build_runtime_url(arn)
        assert url == (
            "https://bedrock-agentcore.us-east-1.amazonaws.com"
            "/runtimes/arn%3Aaws%3Abedrock-agentcore%3Aus-east-1%3A123456789012%3Aruntime%2Fmy-agent-abc123"
            "/invocations"
        )

    def test_region_override(self):
        arn = "arn:aws:bedrock-agentcore:us-east-1:123456789012:runtime/my-agent"
        url = build_runtime_url(arn, region="eu-west-1")
        assert "bedrock-agentcore.eu-west-1.amazonaws.com" in url

    def test_invalid_arn_raises(self):
        import pytest

        with pytest.raises(ValueError, match="Cannot extract region"):
            build_runtime_url("not-an-arn")



"""Shared protocol and utilities for memory converters."""

from typing import Any, Protocol, Tuple

from strands.types.session import SessionMessage

CONVERSATIONAL_MAX_SIZE = 100000


class MemoryConverter(Protocol):
    """Protocol for converting between Strands messages and STM event payloads."""

    @staticmethod
    def message_to_payload(session_message: SessionMessage) -> list[Tuple[str, str]]:
        """Convert SessionMessage to STM event payload format."""

    @staticmethod
    def events_to_messages(events: list[dict[str, Any]]) -> list[SessionMessage]:
        """Convert STM events to SessionMessages."""

    @staticmethod
    def exceeds_conversational_limit(message: tuple[str, str]) -> bool:
        """Check if message exceeds conversational payload size limit."""


def exceeds_conversational_limit(message: tuple[str, str]) -> bool:
    """Check if message exceeds the conversational payload size limit."""
    return sum(len(text) for text in message) >= CONVERSATIONAL_MAX_SIZE



"""Utilities for wrapping boto3 methods to accept snake_case kwargs."""

import functools
import re
from typing import Any, Callable, Dict

_VALID_SNAKE_RE = re.compile(r"^[a-z][a-z0-9]*(_[a-z0-9]+)*$")


def snake_to_camel(name: str) -> str:
    """Convert a snake_case string to camelCase.

    Already-camelCase strings pass through unchanged (no underscores to split on).
    Raises ValueError for malformed snake_case (e.g. leading/trailing underscores,
    consecutive underscores, uppercase characters).
    """
    if "_" not in name:
        return name
    if not _VALID_SNAKE_RE.match(name):
        raise ValueError(f"Invalid parameter name: '{name}'")
    parts = name.split("_")
    return parts[0] + "".join(p.title() for p in parts[1:])


def accept_snake_case_kwargs(method: Callable[..., Any]) -> Callable[..., Any]:
    """Wrap a boto3 method to accept both snake_case and camelCase kwargs.

    Converts all snake_case kwargs to camelCase before forwarding.
    Raises TypeError if both forms are provided (e.g. memory_id and memoryId).
    """

    @functools.wraps(method)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        converted: Dict[str, Any] = {}
        original_keys: Dict[str, str] = {}
        for key, value in kwargs.items():
            camel_key = snake_to_camel(key)
            if camel_key in converted:
                raise TypeError(
                    f"Got both '{original_keys[camel_key]}' and '{key}' for the same parameter. Use one or the other."
                )
            original_keys[camel_key] = key
            converted[camel_key] = value
        return method(*args, **converted)

    return wrapper



"""Bedrock AgentCore Memory module for agent memory management capabilities."""

from .client import MemoryClient
from .controlplane import MemoryControlPlaneClient
from .session import Actor, MemorySession, MemorySessionManager

__all__ = ["Actor", "MemoryClient", "MemorySession", "MemorySessionManager", "MemoryControlPlaneClient"]



"""Evaluation utilities."""

from bedrock_agentcore.evaluation.span_to_adot_serializer import (
    convert_strands_to_adot,
)
from bedrock_agentcore.evaluation.utils.cloudwatch_span_helper import (
    CloudWatchSpanHelper,
    fetch_spans_from_cloudwatch,
)

__all__ = [
    "CloudWatchSpanHelper",
    "fetch_spans_from_cloudwatch",
    "convert_strands_to_adot",
]



"""Tests for manual async task management and edge case coverage."""

import asyncio
import json
import time
from unittest.mock import Mock, patch

import pytest

from bedrock_agentcore.runtime import BedrockAgentCoreApp
from bedrock_agentcore.runtime.models import PingStatus


class TestManualAsyncTaskManagement:
    """Test manual async task management functionality."""

    def test_add_async_task_with_metadata(self):
        """Test add_async_task with metadata parameter."""
        app = BedrockAgentCoreApp()

        # Test with metadata
        metadata = {"file": "data.csv", "priority": "high"}
        task_id = app.add_async_task("file_processing", metadata)

        assert isinstance(task_id, int)
        assert len(app._active_tasks) == 1

        # Verify metadata is stored
        task_info = app._active_tasks[task_id]
        assert task_info["name"] == "file_processing"
        assert task_info["metadata"] == metadata
        assert "start_time" in task_info

    def test_add_async_task_without_metadata(self):
        """Test add_async_task without metadata parameter."""
        app = BedrockAgentCoreApp()

        task_id = app.add_async_task("simple_task")

        assert isinstance(task_id, int)
        assert len(app._active_tasks) == 1

        # Verify no metadata key when not provided
        task_info = app._active_tasks[task_id]
        assert task_info["name"] == "simple_task"
        assert "metadata" not in task_info

    def test_complete_unknown_task_id(self):
        """Test completing a task ID that doesn't exist."""
        app = BedrockAgentCoreApp()

        # Try to complete non-existent task
        result = app.complete_async_task(999999)

        assert result is False
        assert len(app._active_tasks) == 0

    def test_complete_async_task_success(self):
        """Test successful task completion."""
        app = BedrockAgentCoreApp()

        task_id = app.add_async_task("test_task")
        assert len(app._active_tasks) == 1

        result = app.complete_async_task(task_id)

        assert result is True
        assert len(app._active_tasks) == 0

    def test_get_async_task_info_with_corrupted_data(self):
        """Test get_async_task_info handles corrupted task data gracefully."""
        app = BedrockAgentCoreApp()

        # Add corrupted task data (missing required fields)
        app._active_tasks[1] = {"invalid": "data"}  # Missing name and start_time
        app._active_tasks[2] = {"name": "valid_task", "start_time": time.time()}
        app._active_tasks[3] = {"name": "bad_time", "start_time": "not_a_number"}

        # Should handle corrupted data gracefully
        task_info = app.get_async_task_info()

        assert isinstance(task_info, dict)
        assert "active_count" in task_info
        assert "running_jobs" in task_info
        assert task_info["active_count"] == 3  # All tasks counted

        # Only valid jobs should be in running_jobs
        valid_jobs = [job for job in task_info["running_jobs"] if "name" in job and "duration" in job]
        assert len(valid_jobs) <= 2  # At most 2 valid jobs


class TestErrorHandlingScenarios:
    """Test error handling and exception scenarios."""

    @pytest.mark.asyncio
    async def test_invocation_with_malformed_json(self):
        """Test handling of malformed JSON in invocation requests."""
        app = BedrockAgentCoreApp()

        @app.entrypoint
        def test_handler(event):
            return {"result": "ok"}

        # Mock request with invalid JSON
        class MockBadJSONRequest:
            async def json(self):
                raise json.JSONDecodeError("Invalid JSON", "test", 0)

            headers = {}

        request = MockBadJSONRequest()
        response = await app._handle_invocation(request)

        assert response.status_code == 400

    def test_ping_endpoint_exception_handling(self):
        """Test ping endpoint handles exceptions gracefully."""
        app = BedrockAgentCoreApp()

        # Mock get_current_ping_status to raise exception
        with patch.object(app, "get_current_ping_status", side_effect=RuntimeError("Ping failed")):
            response = app._handle_ping(Mock())

            assert response.status_code == 200  # Should return fallback response

    @pytest.mark.asyncio
    async def test_debug_action_exception_handling(self):
        """Test debug action exception handling."""
        app = BedrockAgentCoreApp(debug=True)

        @app.entrypoint
        def test_handler(event):
            return {"result": "ok"}

        # Mock force_ping_status to raise exception
        with patch.object(app, "force_ping_status", side_effect=RuntimeError("Force failed")):

            class MockRequest:
                async def json(self):
                    return {"_agent_core_app_action": "force_healthy"}

                headers = {}

            response = await app._handle_invocation(MockRequest())
            assert response.status_code == 500

    def test_sse_chunk_normal_serialization(self):
        """Test normal SSE chunk serialization."""
        app = BedrockAgentCoreApp()

        # Test with dict
        data = {"message": "hello", "count": 42}
        result = app._convert_to_sse(data)
        assert result == b'data: {"message": "hello", "count": 42}\n\n'

        # Test with string (now sent as plain text, not JSON-encoded)
        result = app._convert_to_sse("simple string")
        assert result == b'data: "simple string"\n\n'

    def test_custom_ping_handler_result_assignment(self):
        """Test custom ping handler result assignment."""
        app = BedrockAgentCoreApp()

        @app.ping
        def custom_handler():
            return "HealthyBusy"  # String that needs conversion

        status = app.get_current_ping_status()
        assert status == PingStatus.HEALTHY_BUSY


class TestStreamingAndAuthentication:
    """Test streaming responses and authentication handling."""

    @pytest.mark.asyncio
    async def test_streaming_generator_response(self):
        """Test streaming response with generator."""
        app = BedrockAgentCoreApp()

        def generator_handler(event):
            yield {"chunk": 1}
            yield {"chunk": 2}
            yield {"chunk": 3}

        @app.entrypoint
        def test_handler(event):
            return generator_handler(event)

        class MockRequest:
            async def json(self):
                return {"test": "data"}

            headers = {}

        response = await app._handle_invocation(MockRequest())

        # Should return StreamingResponse
        assert hasattr(response, "media_type")
        assert response.media_type == "text/event-stream"

    @pytest.mark.asyncio
    async def test_streaming_async_generator_response(self):
        """Test streaming response with async generator."""
        app = BedrockAgentCoreApp()

        async def async_generator_handler(event):
            yield {"chunk": 1}
            yield {"chunk": 2}
            yield {"chunk": 3}

        @app.entrypoint
        async def test_handler(event):
            return async_generator_handler(event)

        class MockRequest:
            async def json(self):
                return {"test": "data"}

            headers = {}

        response = await app._handle_invocation(MockRequest())

        # Should return StreamingResponse
        assert hasattr(response, "media_type")
        assert response.media_type == "text/event-stream"

    @pytest.mark.asyncio
    async def test_authentication_token_handling(self):
        """Test authentication token setting."""
        app = BedrockAgentCoreApp()

        @app.entrypoint
        def test_handler(event, context):
            # Return context to verify it was set
            return {"context_set": context is not None}

        class MockRequest:
            async def json(self):
                return {"test": "data"}

            headers = {"X-Agent-Access-Token": "test-token-123"}

        # Test that handler with context parameter gets called
        response = await app._handle_invocation(MockRequest())
        assert response.status_code == 200

        # Test authentication token extraction
        token = MockRequest().headers.get("X-Agent-Access-Token")
        assert token == "test-token-123"

    @pytest.mark.asyncio
    async def test_no_task_action_return_path(self):
        """Test task action return path when no action is present."""
        app = BedrockAgentCoreApp(debug=True)

        @app.entrypoint
        def test_handler(event):
            return {"result": "ok"}

        class MockRequest:
            async def json(self):
                return {"normal": "request"}  # No _agent_core_app_action

            headers = {}

        # Should return None from _handle_task_action and proceed normally
        response = await app._handle_invocation(MockRequest())
        assert response.status_code == 200


class TestIntegrationScenarios:
    """Test integration scenarios with multiple features."""

    def test_mixed_manual_and_decorator_tasks(self):
        """Test mixing manual task management with decorator tasks."""
        app = BedrockAgentCoreApp()

        @app.async_task
        async def decorated_task():
            await asyncio.sleep(0.01)
            return "decorated_done"

        # Add manual task
        manual_task_id = app.add_async_task("manual_task", {"type": "manual"})

        # Should have one manual task
        assert len(app._active_tasks) == 1
        assert app.get_current_ping_status() == PingStatus.HEALTHY_BUSY

        # Complete manual task
        app.complete_async_task(manual_task_id)
        assert len(app._active_tasks) == 0
        assert app.get_current_ping_status() == PingStatus.HEALTHY

    @pytest.mark.asyncio
    async def test_concurrent_task_management(self):
        """Test concurrent manual task operations."""
        app = BedrockAgentCoreApp()

        # Add multiple tasks concurrently (simulated)
        task_ids = []
        for i in range(5):
            task_id = app.add_async_task(f"task_{i}", {"index": i})
            task_ids.append(task_id)

        assert len(app._active_tasks) == 5
        assert app.get_current_ping_status() == PingStatus.HEALTHY_BUSY

        # Complete tasks
        for task_id in task_ids:
            result = app.complete_async_task(task_id)
            assert result is True

        assert len(app._active_tasks) == 0
        assert app.get_current_ping_status() == PingStatus.HEALTHY

    def test_task_id_uniqueness(self):
        """Test that task IDs are unique."""
        app = BedrockAgentCoreApp()

        task_ids = set()
        for i in range(100):
            task_id = app.add_async_task(f"task_{i}")
            assert task_id not in task_ids
            task_ids.add(task_id)

        # All task IDs should be unique
        assert len(task_ids) == 100
        assert len(app._active_tasks) == 100

    def test_task_lifecycle_logging(self):
        """Test that task lifecycle generates appropriate log messages."""
        app = BedrockAgentCoreApp()

        with patch.object(app.logger, "info") as mock_info:
            # Add task
            task_id = app.add_async_task("logged_task")

            # Complete task
            app.complete_async_task(task_id)

            # Verify logging calls
            assert mock_info.call_count >= 2  # At least start and complete messages

    @pytest.mark.asyncio
    async def test_error_resilience_with_active_tasks(self):
        """Test system resilience when errors occur with active tasks."""
        app = BedrockAgentCoreApp()

        # Add some tasks
        task_id1 = app.add_async_task("task1")
        task_id2 = app.add_async_task("task2")

        # Corrupt one task's data
        app._active_tasks[task_id1] = {"corrupted": "data"}

        # System should still function
        ping_status = app.get_current_ping_status()
        assert ping_status == PingStatus.HEALTHY_BUSY

        task_info = app.get_async_task_info()
        assert task_info["active_count"] == 2

        # Clean completion should still work for valid tasks
        result = app.complete_async_task(task_id2)
        assert result is True


class TestEdgeCasesAndBoundaryConditions:
    """Test edge cases and boundary conditions."""

    def test_task_completion_race_condition_simulation(self):
        """Test task completion under simulated race conditions."""
        app = BedrockAgentCoreApp()

        task_id = app.add_async_task("race_task")

        # Simulate race condition by completing twice
        result1 = app.complete_async_task(task_id)
        result2 = app.complete_async_task(task_id)

        assert result1 is True  # First completion succeeds
        assert result2 is False  # Second completion fails

    def test_large_metadata_handling(self):
        """Test handling of large metadata objects."""
        app = BedrockAgentCoreApp()

        # Create large metadata
        large_metadata = {f"key_{i}": f"value_{i}" * 100 for i in range(100)}

        task_id = app.add_async_task("large_meta_task", large_metadata)

        # Should handle large metadata without issues
        task_info = app._active_tasks[task_id]
        assert task_info["metadata"] == large_metadata

        # Cleanup
        app.complete_async_task(task_id)

    def test_task_duration_calculation_accuracy(self):
        """Test accuracy of task duration calculations."""
        app = BedrockAgentCoreApp()
        task_id = app.add_async_task("duration_test")

        # Wait a bit
        time.sleep(0.1)

        task_info = app.get_async_task_info()
        job = task_info["running_jobs"][0]

        expected_min_duration = 0.05  # At least 50ms
        assert job["duration"] >= expected_min_duration

        app.complete_async_task(task_id)

    @pytest.mark.asyncio
    async def test_context_parameter_detection(self):
        """Test detection of context parameter in handlers."""
        app = BedrockAgentCoreApp()

        @app.entrypoint
        def handler_with_context(event, context):
            return {"has_context": True}

        @app.entrypoint
        def handler_without_context(event):
            return {"has_context": False}

        # Test with context handler
        app.handlers["main"] = handler_with_context
        assert app._takes_context(handler_with_context) is True

        # Test without context handler
        app.handlers["main"] = handler_without_context
        assert app._takes_context(handler_without_context) is False


if __name__ == "__main__":
    pytest.main([__file__, "-v"])



"""
Integration tests for Strands AgentCore Evaluation.

Run with: python -m pytest tests_integ/evaluation/integrations/strands/test_strands_evaluation.py -v
"""

import logging
import os

import pytest
from strands import Agent, tool
from strands_evals import Case, Experiment
from strands_evals.telemetry import StrandsEvalsTelemetry

from bedrock_agentcore.evaluation import create_strands_evaluator

logging.basicConfig(level=logging.INFO, format="%(asctime)s - %(levelname)s - %(message)s")
logger = logging.getLogger(__name__)

REGION = os.environ.get("BEDROCK_TEST_REGION", "us-west-2")

# Suppress Pydantic serialization warnings for OTel spans
pytestmark = pytest.mark.filterwarnings("ignore::UserWarning:pydantic.main")


@tool
def calculator(expression: str) -> str:
    """Evaluates a mathematical expression."""
    try:
        return str(eval(expression))
    except Exception as e:
        return f"Error: {str(e)}"


@pytest.mark.integration
class TestStrandsEvaluationIntegration:
    """Real integration tests for Strands AgentCore Evaluation."""

    @classmethod
    def setup_class(cls):
        """Set up test environment."""
        cls.region = os.environ.get("BEDROCK_TEST_REGION", "us-west-2")

    def test_real_evaluation_with_builtin_helpfulness(self):
        """Test real evaluation with Builtin.Helpfulness evaluator."""
        telemetry = StrandsEvalsTelemetry().setup_in_memory_exporter()

        def task_fn(case):
            agent = Agent(
                tools=[calculator],
                system_prompt="You are a helpful math assistant. Use the calculator tool to solve problems.",
            )
            agent_response = agent(case.input)

            # Convert tuple to list to avoid Pydantic warning
            raw_spans = list(telemetry.in_memory_exporter.get_finished_spans())
            return {"output": str(agent_response), "trajectory": raw_spans}

        cases = [Case(input="What is 2+2?", expected_output="4")]

        evaluator = create_strands_evaluator("Builtin.Helpfulness", region=REGION)
        experiment = Experiment(cases=cases, evaluators=[evaluator])
        reports = experiment.run_evaluations(task_fn)
        report = reports[0]

        # Verify results
        assert report.overall_score >= 0.0
        assert report.overall_score <= 1.0
        logger.info("Evaluation score: %s", report.overall_score)

    def test_real_evaluation_with_builtin_accuracy(self):
        """Test real evaluation with Builtin.Accuracy evaluator."""
        telemetry = StrandsEvalsTelemetry().setup_in_memory_exporter()

        def task_fn(case):
            agent = Agent(
                tools=[calculator],
                system_prompt="You are a helpful math assistant. Use the calculator tool to solve problems.",
            )
            agent_response = agent(case.input)

            raw_spans = list(telemetry.in_memory_exporter.get_finished_spans())
            return {"output": str(agent_response), "trajectory": raw_spans}

        cases = [Case(input="Calculate 5 + 3", expected_output="8")]

        evaluator = create_strands_evaluator("Builtin.Accuracy", region=REGION)
        experiment = Experiment(cases=cases, evaluators=[evaluator])
        reports = experiment.run_evaluations(task_fn)
        report = reports[0]

        assert report.overall_score >= 0.0
        assert report.overall_score <= 1.0
        logger.info("Accuracy score: %s", report.overall_score)

    def test_real_evaluation_with_multiple_cases(self):
        """Test real evaluation with multiple test cases."""
        telemetry = StrandsEvalsTelemetry().setup_in_memory_exporter()

        def task_fn(case):
            agent = Agent(
                tools=[calculator],
                system_prompt="You are a helpful math assistant. Use the calculator tool to solve problems.",
            )
            agent_response = agent(case.input)

            raw_spans = list(telemetry.in_memory_exporter.get_finished_spans())
            return {"output": str(agent_response), "trajectory": raw_spans}

        cases = [
            Case(input="What is 5 + 3?", expected_output="8"),
            Case(input="Calculate 10 + 7", expected_output="17"),
            Case(input="What is 100 - 25?", expected_output="75"),
        ]

        evaluator = create_strands_evaluator("Builtin.Helpfulness", region=REGION, test_pass_score=0.6)
        experiment = Experiment(cases=cases, evaluators=[evaluator])
        reports = experiment.run_evaluations(task_fn)
        report = reports[0]

        assert report.overall_score >= 0.0
        assert report.overall_score <= 1.0
        assert len(report.test_passes) == 3
        pass_rate = sum(report.test_passes) / len(report.test_passes)
        logger.info("Average score: %.2f", report.overall_score)
        logger.info("Pass rate: %.1f%%", pass_rate * 100)

    def test_evaluation_with_empty_trajectory(self):
        """Test evaluation handles empty trajectory gracefully."""

        def task_fn(case):
            return {"output": "Response", "trajectory": []}

        cases = [Case(input="Test", expected_output="Response")]

        evaluator = create_strands_evaluator("Builtin.Helpfulness", region=REGION)
        experiment = Experiment(cases=cases, evaluators=[evaluator])
        reports = experiment.run_evaluations(task_fn)
        report = reports[0]

        # Should return 0 score for empty trajectory
        assert report.overall_score == 0.0
        assert not any(report.test_passes)

    def test_evaluation_with_custom_pass_score(self):
        """Test evaluation with custom test pass score threshold."""
        telemetry = StrandsEvalsTelemetry().setup_in_memory_exporter()

        def task_fn(case):
            agent = Agent(
                tools=[calculator],
                system_prompt="You are a helpful math assistant. Use the calculator tool to solve problems.",
            )
            agent_response = agent(case.input)

            raw_spans = list(telemetry.in_memory_exporter.get_finished_spans())
            return {"output": str(agent_response), "trajectory": raw_spans}

        cases = [Case(input="What is 2+2?", expected_output="4")]

        # Test with high threshold
        evaluator = create_strands_evaluator("Builtin.Helpfulness", region=REGION, test_pass_score=0.9)
        experiment = Experiment(cases=cases, evaluators=[evaluator])
        reports = experiment.run_evaluations(task_fn)
        report = reports[0]

        assert report.overall_score >= 0.0
        assert report.overall_score <= 1.0
        logger.info("Score with 0.9 threshold: %s", report.overall_score)



"""Tests for AgentCoreRuntimeClient."""

from unittest.mock import Mock, patch
from urllib.parse import quote

import pytest

from bedrock_agentcore.runtime.agent_core_runtime_client import AgentCoreRuntimeClient


class TestAgentCoreRuntimeClientInit:
    """Tests for AgentCoreRuntimeClient initialization."""

    def test_init_stores_region(self):
        """Test that initialization stores the region."""
        client = AgentCoreRuntimeClient(region="us-west-2")
        assert client.region == "us-west-2"

    def test_init_creates_logger(self):
        """Test that initialization creates a logger."""
        client = AgentCoreRuntimeClient(region="us-west-2")
        assert client.logger is not None


class TestParseRuntimeArn:
    """Tests for _parse_runtime_arn helper."""

    def test_parse_valid_arn(self):
        """Test parsing a valid runtime ARN."""
        client = AgentCoreRuntimeClient(region="us-west-2")
        arn = "arn:aws:bedrock-agentcore:us-west-2:123456789012:runtime/my-runtime-abc123"

        result = client._parse_runtime_arn(arn)

        assert result["region"] == "us-west-2"
        assert result["account_id"] == "123456789012"
        assert result["runtime_id"] == "my-runtime-abc123"

    def test_parse_invalid_arn_raises_error(self):
        """Test that invalid ARN format raises ValueError."""
        client = AgentCoreRuntimeClient(region="us-west-2")
        invalid_arn = "not-a-valid-arn"

        with pytest.raises(ValueError, match="Invalid runtime ARN format"):
            client._parse_runtime_arn(invalid_arn)

    def test_parse_wrong_service_raises_error(self):
        """Test that wrong service in ARN raises ValueError."""
        client = AgentCoreRuntimeClient(region="us-west-2")
        wrong_service = "arn:aws:s3:us-west-2:123456789012:bucket/my-bucket"

        with pytest.raises(ValueError, match="Invalid runtime ARN format"):
            client._parse_runtime_arn(wrong_service)

    def test_parse_empty_region_raises_error(self):
        """Test that empty region in ARN raises ValueError."""
        client = AgentCoreRuntimeClient(region="us-west-2")
        empty_region = "arn:aws:bedrock-agentcore::123456789012:runtime/my-runtime"

        with pytest.raises(ValueError, match="ARN components cannot be empty"):
            client._parse_runtime_arn(empty_region)

    def test_parse_empty_account_id_raises_error(self):
        """Test that empty account_id in ARN raises ValueError."""
        client = AgentCoreRuntimeClient(region="us-west-2")
        empty_account = "arn:aws:bedrock-agentcore:us-west-2::runtime/my-runtime"

        with pytest.raises(ValueError, match="ARN components cannot be empty"):
            client._parse_runtime_arn(empty_account)

    def test_parse_empty_runtime_id_raises_error(self):
        """Test that empty runtime_id in ARN raises ValueError."""
        client = AgentCoreRuntimeClient(region="us-west-2")
        empty_runtime = "arn:aws:bedrock-agentcore:us-west-2:123456789012:runtime/"

        with pytest.raises(ValueError, match="ARN components cannot be empty"):
            client._parse_runtime_arn(empty_runtime)


class TestBuildWebsocketUrl:
    """Tests for _build_websocket_url helper."""

    @patch("bedrock_agentcore.runtime.agent_core_runtime_client.get_data_plane_endpoint")
    def test_build_basic_url(self, mock_endpoint):
        """Test building basic WebSocket URL without query params."""
        mock_endpoint.return_value = "https://example.aws.dev"
        client = AgentCoreRuntimeClient(region="us-west-2")
        runtime_arn = "arn:aws:bedrock-agentcore:us-west-2:123:runtime/my-runtime"

        result = client._build_websocket_url(runtime_arn)

        # ARN should be URL encoded
        encoded_arn = quote(runtime_arn, safe="")
        assert result == f"wss://example.aws.dev/runtimes/{encoded_arn}/ws"

    @patch("bedrock_agentcore.runtime.agent_core_runtime_client.get_data_plane_endpoint")
    def test_build_url_with_endpoint_name(self, mock_endpoint):
        """Test building URL with endpoint name (qualifier param)."""
        mock_endpoint.return_value = "https://example.aws.dev"
        client = AgentCoreRuntimeClient(region="us-west-2")
        runtime_arn = "arn:aws:bedrock-agentcore:us-west-2:123:runtime/my-runtime"

        result = client._build_websocket_url(runtime_arn, endpoint_name="DEFAULT")

        encoded_arn = quote(runtime_arn, safe="")
        assert result == f"wss://example.aws.dev/runtimes/{encoded_arn}/ws?qualifier=DEFAULT"

    @patch("bedrock_agentcore.runtime.agent_core_runtime_client.get_data_plane_endpoint")
    def test_build_url_with_custom_headers(self, mock_endpoint):
        """Test building URL with custom headers as query params."""
        mock_endpoint.return_value = "https://example.aws.dev"
        client = AgentCoreRuntimeClient(region="us-west-2")
        runtime_arn = "arn:aws:bedrock-agentcore:us-west-2:123:runtime/my-runtime"

        result = client._build_websocket_url(runtime_arn, custom_headers={"abc": "pqr", "foo": "bar"})

        encoded_arn = quote(runtime_arn, safe="")
        assert f"wss://example.aws.dev/runtimes/{encoded_arn}/ws?" in result
        assert "abc=pqr" in result
        assert "foo=bar" in result

    @patch("bedrock_agentcore.runtime.agent_core_runtime_client.get_data_plane_endpoint")
    def test_build_url_with_all_params(self, mock_endpoint):
        """Test building URL with endpoint name and custom headers."""
        mock_endpoint.return_value = "https://example.aws.dev"
        client = AgentCoreRuntimeClient(region="us-west-2")
        runtime_arn = "arn:aws:bedrock-agentcore:us-west-2:123:runtime/my-runtime"

        result = client._build_websocket_url(runtime_arn, endpoint_name="DEFAULT", custom_headers={"abc": "pqr"})

        encoded_arn = quote(runtime_arn, safe="")
        assert f"wss://example.aws.dev/runtimes/{encoded_arn}/ws?" in result
        assert "qualifier=DEFAULT" in result
        assert "abc=pqr" in result


class TestGenerateWsConnection:
    """Tests for generate_ws_connection method."""

    @patch("bedrock_agentcore.runtime.agent_core_runtime_client.boto3.Session")
    @patch("bedrock_agentcore.runtime.agent_core_runtime_client.get_data_plane_endpoint")
    def test_generate_basic_connection(self, mock_endpoint, mock_session):
        """Test generating basic WebSocket connection."""
        # Setup mocks
        mock_endpoint.return_value = "https://example.aws.dev"
        mock_creds = Mock()
        mock_creds.get_frozen_credentials.return_value = Mock(access_key="AKIATEST", secret_key="secret", token=None)
        mock_session.return_value.get_credentials.return_value = mock_creds

        client = AgentCoreRuntimeClient(region="us-west-2")
        runtime_arn = "arn:aws:bedrock-agentcore:us-west-2:123:runtime/my-runtime"

        ws_url, headers = client.generate_ws_connection(runtime_arn)

        # Verify URL structure
        assert ws_url.startswith("wss://example.aws.dev/runtimes/")
        assert "/ws" in ws_url

        # Verify required headers
        assert "Host" in headers
        assert "X-Amz-Date" in headers
        assert "Authorization" in headers
        assert "Upgrade" in headers
        assert "Connection" in headers
        assert "Sec-WebSocket-Version" in headers
        assert "Sec-WebSocket-Key" in headers

    @patch("bedrock_agentcore.runtime.agent_core_runtime_client.boto3.Session")
    @patch("bedrock_agentcore.runtime.agent_core_runtime_client.get_data_plane_endpoint")
    def test_generate_connection_with_session_id(self, mock_endpoint, mock_session):
        """Test generating connection with explicit session ID."""
        mock_endpoint.return_value = "https://example.aws.dev"
        mock_creds = Mock()
        mock_creds.get_frozen_credentials.return_value = Mock(access_key="AKIATEST", secret_key="secret", token=None)
        mock_session.return_value.get_credentials.return_value = mock_creds

        client = AgentCoreRuntimeClient(region="us-west-2")
        runtime_arn = "arn:aws:bedrock-agentcore:us-west-2:123:runtime/my-runtime"

        ws_url, headers = client.generate_ws_connection(runtime_arn, session_id="test-session-123")

        assert ws_url is not None
        assert headers is not None
        # Verify session ID is in headers
        assert "X-Amzn-Bedrock-AgentCore-Runtime-Session-Id" in headers
        assert headers["X-Amzn-Bedrock-AgentCore-Runtime-Session-Id"] == "test-session-123"

    @patch("bedrock_agentcore.runtime.agent_core_runtime_client.boto3.Session")
    @patch("bedrock_agentcore.runtime.agent_core_runtime_client.get_data_plane_endpoint")
    def test_generate_connection_user_agent(self, mock_endpoint, mock_session):
        """Test that User-Agent header is set correctly."""
        mock_endpoint.return_value = "https://example.aws.dev"
        mock_creds = Mock()
        mock_creds.get_frozen_credentials.return_value = Mock(access_key="AKIATEST", secret_key="secret", token=None)
        mock_session.return_value.get_credentials.return_value = mock_creds

        client = AgentCoreRuntimeClient(region="us-west-2")
        runtime_arn = "arn:aws:bedrock-agentcore:us-west-2:123:runtime/my-runtime"

        ws_url, headers = client.generate_ws_connection(runtime_arn)

        assert "User-Agent" in headers
        assert headers["User-Agent"] == "AgentCoreRuntimeClient/1.0"

    @patch("bedrock_agentcore.runtime.agent_core_runtime_client.boto3.Session")
    @patch("bedrock_agentcore.runtime.agent_core_runtime_client.get_data_plane_endpoint")
    def test_generate_connection_with_endpoint_name(self, mock_endpoint, mock_session):
        """Test generating connection with endpoint name."""
        mock_endpoint.return_value = "https://example.aws.dev"
        mock_creds = Mock()
        mock_creds.get_frozen_credentials.return_value = Mock(access_key="AKIATEST", secret_key="secret", token=None)
        mock_session.return_value.get_credentials.return_value = mock_creds

        client = AgentCoreRuntimeClient(region="us-west-2")
        runtime_arn = "arn:aws:bedrock-agentcore:us-west-2:123:runtime/my-runtime"

        ws_url, headers = client.generate_ws_connection(runtime_arn, endpoint_name="DEFAULT")

        assert "qualifier=DEFAULT" in ws_url

    @patch("bedrock_agentcore.runtime.agent_core_runtime_client.boto3.Session")
    def test_generate_connection_no_credentials_raises_error(self, mock_session):
        """Test that missing credentials raises RuntimeError."""
        mock_session.return_value.get_credentials.return_value = None

        client = AgentCoreRuntimeClient(region="us-west-2")
        runtime_arn = "arn:aws:bedrock-agentcore:us-west-2:123:runtime/my-runtime"

        with pytest.raises(RuntimeError, match="No AWS credentials found"):
            client.generate_ws_connection(runtime_arn)


class TestGeneratePresignedUrl:
    """Tests for generate_presigned_url method."""

    @patch("bedrock_agentcore.runtime.agent_core_runtime_client.boto3.Session")
    @patch("bedrock_agentcore.runtime.agent_core_runtime_client.get_data_plane_endpoint")
    def test_generate_basic_presigned_url(self, mock_endpoint, mock_session):
        """Test generating basic presigned URL."""
        mock_endpoint.return_value = "https://example.aws.dev"
        mock_creds = Mock()
        mock_creds.get_frozen_credentials.return_value = Mock(access_key="AKIATEST", secret_key="secret", token=None)
        mock_session.return_value.get_credentials.return_value = mock_creds

        client = AgentCoreRuntimeClient(region="us-west-2")
        runtime_arn = "arn:aws:bedrock-agentcore:us-west-2:123:runtime/my-runtime"

        presigned_url = client.generate_presigned_url(runtime_arn)

        # Verify URL structure
        assert presigned_url.startswith("wss://example.aws.dev/runtimes/")
        assert "/ws?" in presigned_url

        # Verify SigV4 query parameters
        assert "X-Amz-Algorithm" in presigned_url
        assert "X-Amz-Credential" in presigned_url
        assert "X-Amz-Date" in presigned_url
        assert "X-Amz-Expires" in presigned_url
        assert "X-Amz-SignedHeaders" in presigned_url
        assert "X-Amz-Signature" in presigned_url

    @patch("bedrock_agentcore.runtime.agent_core_runtime_client.boto3.Session")
    @patch("bedrock_agentcore.runtime.agent_core_runtime_client.get_data_plane_endpoint")
    def test_generate_presigned_url_with_endpoint_name(self, mock_endpoint, mock_session):
        """Test generating presigned URL with endpoint name."""
        mock_endpoint.return_value = "https://example.aws.dev"
        mock_creds = Mock()
        mock_creds.get_frozen_credentials.return_value = Mock(access_key="AKIATEST", secret_key="secret", token=None)
        mock_session.return_value.get_credentials.return_value = mock_creds

        client = AgentCoreRuntimeClient(region="us-west-2")
        runtime_arn = "arn:aws:bedrock-agentcore:us-west-2:123:runtime/my-runtime"

        presigned_url = client.generate_presigned_url(runtime_arn, endpoint_name="DEFAULT")

        assert "qualifier=DEFAULT" in presigned_url

    @patch("bedrock_agentcore.runtime.agent_core_runtime_client.boto3.Session")
    @patch("bedrock_agentcore.runtime.agent_core_runtime_client.get_data_plane_endpoint")
    def test_generate_presigned_url_with_custom_headers(self, mock_endpoint, mock_session):
        """Test generating presigned URL with custom headers."""
        mock_endpoint.return_value = "https://example.aws.dev"
        mock_creds = Mock()
        mock_creds.get_frozen_credentials.return_value = Mock(access_key="AKIATEST", secret_key="secret", token=None)
        mock_session.return_value.get_credentials.return_value = mock_creds

        client = AgentCoreRuntimeClient(region="us-west-2")
        runtime_arn = "arn:aws:bedrock-agentcore:us-west-2:123:runtime/my-runtime"

        presigned_url = client.generate_presigned_url(runtime_arn, custom_headers={"abc": "pqr"})

        assert "abc=pqr" in presigned_url

    @patch("bedrock_agentcore.runtime.agent_core_runtime_client.boto3.Session")
    @patch("bedrock_agentcore.runtime.agent_core_runtime_client.get_data_plane_endpoint")
    def test_generate_presigned_url_with_session_id(self, mock_endpoint, mock_session):
        """Test generating presigned URL with explicit session ID."""
        mock_endpoint.return_value = "https://example.aws.dev"
        mock_creds = Mock()
        mock_creds.get_frozen_credentials.return_value = Mock(access_key="AKIATEST", secret_key="secret", token=None)
        mock_session.return_value.get_credentials.return_value = mock_creds

        client = AgentCoreRuntimeClient(region="us-west-2")
        runtime_arn = "arn:aws:bedrock-agentcore:us-west-2:123:runtime/my-runtime"

        presigned_url = client.generate_presigned_url(runtime_arn, session_id="test-session-456")

        # Verify session ID is in query params
        assert "X-Amzn-Bedrock-AgentCore-Runtime-Session-Id=test-session-456" in presigned_url

    @patch("bedrock_agentcore.runtime.agent_core_runtime_client.boto3.Session")
    @patch("bedrock_agentcore.runtime.agent_core_runtime_client.get_data_plane_endpoint")
    def test_generate_presigned_url_with_custom_expires(self, mock_endpoint, mock_session):
        """Test generating presigned URL with custom expiration."""
        mock_endpoint.return_value = "https://example.aws.dev"
        mock_creds = Mock()
        mock_creds.get_frozen_credentials.return_value = Mock(access_key="AKIATEST", secret_key="secret", token=None)
        mock_session.return_value.get_credentials.return_value = mock_creds

        client = AgentCoreRuntimeClient(region="us-west-2")
        runtime_arn = "arn:aws:bedrock-agentcore:us-west-2:123:runtime/my-runtime"

        presigned_url = client.generate_presigned_url(runtime_arn, expires=60)

        assert "X-Amz-Expires=60" in presigned_url
        # Verify auto-generated session ID is present
        assert "X-Amzn-Bedrock-AgentCore-Runtime-Session-Id=" in presigned_url

    def test_generate_presigned_url_exceeds_max_expires_raises_error(self):
        """Test that exceeding max expiration raises ValueError."""
        client = AgentCoreRuntimeClient(region="us-west-2")
        runtime_arn = "arn:aws:bedrock-agentcore:us-west-2:123:runtime/my-runtime"

        with pytest.raises(ValueError, match="Expiry timeout cannot exceed"):
            client.generate_presigned_url(runtime_arn, expires=400)

    @patch("bedrock_agentcore.runtime.agent_core_runtime_client.boto3.Session")
    def test_generate_presigned_url_no_credentials_raises_error(self, mock_session):
        """Test that missing credentials raises RuntimeError."""
        mock_session.return_value.get_credentials.return_value = None

        client = AgentCoreRuntimeClient(region="us-west-2")
        runtime_arn = "arn:aws:bedrock-agentcore:us-west-2:123:runtime/my-runtime"

        with pytest.raises(RuntimeError, match="No AWS credentials found"):
            client.generate_presigned_url(runtime_arn)


class TestAgentCoreRuntimeClientSession:
    """Tests for AgentCoreRuntimeClient with custom boto3 session."""

    def test_init_with_custom_session(self):
        """Test initialization with custom boto3 session."""
        custom_session = Mock()
        client = AgentCoreRuntimeClient(region="us-west-2", session=custom_session)

        assert client.region == "us-west-2"
        assert client.session == custom_session

    @patch("bedrock_agentcore.runtime.agent_core_runtime_client.boto3.Session")
    def test_init_without_session_creates_default(self, mock_session_class):
        """Test that default session is created when not provided."""
        mock_session = Mock()
        mock_session_class.return_value = mock_session

        client = AgentCoreRuntimeClient(region="us-west-2")

        assert client.session == mock_session
        mock_session_class.assert_called_once()

    @patch("bedrock_agentcore.runtime.agent_core_runtime_client.get_data_plane_endpoint")
    def test_generate_ws_connection_uses_custom_session(self, mock_endpoint):
        """Test that generate_ws_connection uses the custom session."""
        mock_endpoint.return_value = "https://example.aws.dev"

        # Create custom session with credentials
        custom_session = Mock()
        mock_creds = Mock()
        mock_creds.get_frozen_credentials.return_value = Mock(access_key="AKIATEST", secret_key="secret", token=None)
        custom_session.get_credentials.return_value = mock_creds

        client = AgentCoreRuntimeClient(region="us-west-2", session=custom_session)
        runtime_arn = "arn:aws:bedrock-agentcore:us-west-2:123:runtime/my-runtime"

        ws_url, headers = client.generate_ws_connection(runtime_arn)

        # Verify custom session was used
        custom_session.get_credentials.assert_called_once()
        assert ws_url.startswith("wss://")
        assert "Authorization" in headers


class TestGenerateWsConnectionOAuth:
    """Tests for generate_ws_connection_oauth method."""

    @patch("bedrock_agentcore.runtime.agent_core_runtime_client.get_data_plane_endpoint")
    def test_generate_oauth_connection_basic(self, mock_endpoint):
        """Test generating basic OAuth WebSocket connection."""
        mock_endpoint.return_value = "https://example.aws.dev"
        client = AgentCoreRuntimeClient(region="us-west-2")
        runtime_arn = "arn:aws:bedrock-agentcore:us-west-2:123:runtime/my-runtime"
        bearer_token = "eyJhbGciOiJSUzI1NiIsInR5cCI6IkpXVCJ9.test.token"

        ws_url, headers = client.generate_ws_connection_oauth(runtime_arn, bearer_token)

        # Verify URL structure
        assert ws_url.startswith("wss://example.aws.dev/runtimes/")
        assert "/ws" in ws_url

        # Verify OAuth headers
        assert "Authorization" in headers
        assert headers["Authorization"] == f"Bearer {bearer_token}"
        assert "X-Amzn-Bedrock-AgentCore-Runtime-Session-Id" in headers
        assert "Sec-WebSocket-Key" in headers
        assert "Sec-WebSocket-Version" in headers
        assert headers["Sec-WebSocket-Version"] == "13"

    @patch("bedrock_agentcore.runtime.agent_core_runtime_client.get_data_plane_endpoint")
    def test_generate_oauth_connection_with_session_id(self, mock_endpoint):
        """Test generating OAuth connection with explicit session ID."""
        mock_endpoint.return_value = "https://example.aws.dev"
        client = AgentCoreRuntimeClient(region="us-west-2")
        runtime_arn = "arn:aws:bedrock-agentcore:us-west-2:123:runtime/my-runtime"
        bearer_token = "test-token"
        custom_session_id = "custom-oauth-session-123"

        ws_url, headers = client.generate_ws_connection_oauth(runtime_arn, bearer_token, session_id=custom_session_id)

        assert headers["X-Amzn-Bedrock-AgentCore-Runtime-Session-Id"] == custom_session_id

    @patch("bedrock_agentcore.runtime.agent_core_runtime_client.get_data_plane_endpoint")
    def test_generate_oauth_connection_with_endpoint_name(self, mock_endpoint):
        """Test generating OAuth connection with endpoint name."""
        mock_endpoint.return_value = "https://example.aws.dev"
        client = AgentCoreRuntimeClient(region="us-west-2")
        runtime_arn = "arn:aws:bedrock-agentcore:us-west-2:123:runtime/my-runtime"
        bearer_token = "test-token"

        ws_url, headers = client.generate_ws_connection_oauth(runtime_arn, bearer_token, endpoint_name="DEFAULT")

        assert "qualifier=DEFAULT" in ws_url

    def test_generate_oauth_connection_empty_token_raises_error(self):
        """Test that empty bearer token raises ValueError."""
        client = AgentCoreRuntimeClient(region="us-west-2")
        runtime_arn = "arn:aws:bedrock-agentcore:us-west-2:123:runtime/my-runtime"

        with pytest.raises(ValueError, match="Bearer token cannot be empty"):
            client.generate_ws_connection_oauth(runtime_arn, "")

    def test_generate_oauth_connection_invalid_arn_raises_error(self):
        """Test that invalid ARN raises ValueError."""
        client = AgentCoreRuntimeClient(region="us-west-2")
        invalid_arn = "invalid-arn"
        bearer_token = "test-token"

        with pytest.raises(ValueError, match="Invalid runtime ARN format"):
            client.generate_ws_connection_oauth(invalid_arn, bearer_token)



"""Client for generating WebSocket authentication for AgentCore Runtime.

This module provides a client for generating authentication credentials
for WebSocket connections to AgentCore Runtime endpoints.
"""

import base64
import datetime
import logging
import secrets
import uuid
from typing import Dict, Optional, Tuple
from urllib.parse import quote, urlencode, urlparse

import boto3
from botocore.auth import SigV4Auth, SigV4QueryAuth
from botocore.awsrequest import AWSRequest

from .._utils.endpoints import get_data_plane_endpoint

DEFAULT_PRESIGNED_URL_TIMEOUT = 300
MAX_PRESIGNED_URL_TIMEOUT = 300


class AgentCoreRuntimeClient:
    """Client for generating WebSocket authentication for AgentCore Runtime.

    This client provides authentication credentials for WebSocket connections
    to AgentCore Runtime endpoints, allowing applications to establish
    bidirectional streaming connections with agent runtimes.

    Attributes:
        region (str): The AWS region being used.
        session (boto3.Session): The boto3 session for AWS credentials.
    """

    def __init__(self, region: str, session: Optional[boto3.Session] = None) -> None:
        """Initialize an AgentCoreRuntime client for the specified AWS region.

        Args:
            region (str): The AWS region to use for the AgentCore Runtime service.
            session (Optional[boto3.Session]): Optional boto3 session. If not provided,
                a new session will be created using default credentials.
        """
        self.region = region
        self.logger = logging.getLogger(__name__)

        if session is None:
            session = boto3.Session()

        self.session = session

    def _parse_runtime_arn(self, runtime_arn: str) -> Dict[str, str]:
        """Parse runtime ARN and extract components.

        Args:
            runtime_arn (str): Full runtime ARN

        Returns:
            Dict[str, str]: Dictionary with region, account_id, runtime_id

        Raises:
            ValueError: If ARN format is invalid
        """
        # Expected format: arn:aws:bedrock-agentcore:{region}:{account}:runtime/{runtime_id}
        parts = runtime_arn.split(":")

        if len(parts) != 6:
            raise ValueError(f"Invalid runtime ARN format: {runtime_arn}")

        if parts[0] != "arn" or parts[1] != "aws" or parts[2] != "bedrock-agentcore":
            raise ValueError(f"Invalid runtime ARN format: {runtime_arn}")

        # Parse the resource part (runtime/{runtime_id})
        resource = parts[5]
        if not resource.startswith("runtime/"):
            raise ValueError(f"Invalid runtime ARN format: {runtime_arn}")

        runtime_id = resource.split("/", 1)[1]

        # Validate that components are not empty
        region = parts[3]
        account_id = parts[4]

        if not region or not account_id or not runtime_id:
            raise ValueError("ARN components cannot be empty")

        return {
            "region": region,
            "account_id": account_id,
            "runtime_id": runtime_id,
        }

    def _build_websocket_url(
        self,
        runtime_arn: str,
        endpoint_name: Optional[str] = None,
        custom_headers: Optional[Dict[str, str]] = None,
    ) -> str:
        """Build WebSocket URL with query parameters.

        Args:
            runtime_arn (str): Full runtime ARN
            endpoint_name (Optional[str]): Optional endpoint name for qualifier param
            custom_headers (Optional[Dict[str, str]]): Optional custom query parameters

        Returns:
            str: WebSocket URL with query parameters
        """
        # Get the data plane endpoint
        host = get_data_plane_endpoint(self.region).replace("https://", "")

        # URL-encode the runtime ARN
        encoded_arn = quote(runtime_arn, safe="")

        # Build base path
        path = f"/runtimes/{encoded_arn}/ws"

        # Build query parameters
        query_params = {}

        if endpoint_name:
            query_params["qualifier"] = endpoint_name

        if custom_headers:
            query_params.update(custom_headers)

        # Construct URL
        if query_params:
            query_string = urlencode(query_params)
            ws_url = f"wss://{host}{path}?{query_string}"
        else:
            ws_url = f"wss://{host}{path}"

        return ws_url

    def generate_ws_connection(
        self,
        runtime_arn: str,
        session_id: Optional[str] = None,
        endpoint_name: Optional[str] = None,
    ) -> Tuple[str, Dict[str, str]]:
        """Generate WebSocket URL and SigV4 signed headers for runtime connection.

        Args:
            runtime_arn (str): Full runtime ARN
                (e.g., 'arn:aws:bedrock-agentcore:us-west-2:123:runtime/my-runtime-abc')
            session_id (Optional[str]): Session ID to use. If None, auto-generates a UUID.
            endpoint_name (Optional[str]): Endpoint name to use as 'qualifier' query parameter.
                If provided, adds ?qualifier={endpoint_name} to the URL.

        Returns:
            Tuple[str, Dict[str, str]]: A tuple containing:
                - WebSocket URL (wss://...) with query parameters
                - Headers dictionary with SigV4 signature

        Raises:
            RuntimeError: If no AWS credentials are found.
            ValueError: If runtime_arn format is invalid.

        Example:
            >>> client = AgentCoreRuntimeClient('us-west-2')
            >>> ws_url, headers = client.generate_ws_connection(
            ...     runtime_arn='arn:aws:bedrock-agentcore:us-west-2:123:runtime/my-runtime',
            ...     endpoint_name='DEFAULT'
            ... )
        """
        self.logger.info("Generating WebSocket connection credentials...")

        # Validate ARN
        self._parse_runtime_arn(runtime_arn)

        # Auto-generate session ID if not provided
        if not session_id:
            session_id = str(uuid.uuid4())
            self.logger.debug("Auto-generated session ID: %s", session_id)

        # Build WebSocket URL
        ws_url = self._build_websocket_url(runtime_arn, endpoint_name)

        # Get AWS credentials
        credentials = self.session.get_credentials()
        if not credentials:
            raise RuntimeError("No AWS credentials found")

        frozen_credentials = credentials.get_frozen_credentials()

        # Convert wss:// to https:// for signing
        https_url = ws_url.replace("wss://", "https://")
        parsed = urlparse(https_url)
        host = parsed.netloc

        # Create the request to sign
        request = AWSRequest(
            method="GET",
            url=https_url,
            headers={
                "host": host,
                "x-amz-date": datetime.datetime.now(datetime.timezone.utc).strftime("%Y%m%dT%H%M%SZ"),
            },
        )

        # Sign the request with SigV4
        auth = SigV4Auth(frozen_credentials, "bedrock-agentcore", self.region)
        auth.add_auth(request)

        # Build headers for WebSocket connection
        headers = {
            "Host": host,
            "X-Amz-Date": request.headers["x-amz-date"],
            "Authorization": request.headers["Authorization"],
            "X-Amzn-Bedrock-AgentCore-Runtime-Session-Id": session_id,
            "Upgrade": "websocket",
            "Connection": "Upgrade",
            "Sec-WebSocket-Version": "13",
            "Sec-WebSocket-Key": base64.b64encode(secrets.token_bytes(16)).decode(),
            "User-Agent": "AgentCoreRuntimeClient/1.0",
        }

        # Add session token if present
        if frozen_credentials.token:
            headers["X-Amz-Security-Token"] = frozen_credentials.token

        self.logger.info("✓ WebSocket connection credentials generated (Session: %s)", session_id)
        return ws_url, headers

    def generate_presigned_url(
        self,
        runtime_arn: str,
        session_id: Optional[str] = None,
        endpoint_name: Optional[str] = None,
        custom_headers: Optional[Dict[str, str]] = None,
        expires: int = DEFAULT_PRESIGNED_URL_TIMEOUT,
    ) -> str:
        """Generate a presigned WebSocket URL for runtime connection.

        Presigned URLs include authentication in query parameters, allowing
        frontend clients to connect without managing AWS credentials.

        Args:
            runtime_arn (str): Full runtime ARN
                (e.g., 'arn:aws:bedrock-agentcore:us-west-2:123:runtime/my-runtime-abc')
            session_id (Optional[str]): Session ID to use. If None, auto-generates a UUID.
            endpoint_name (Optional[str]): Endpoint name to use as 'qualifier' query parameter.
                If provided, adds ?qualifier={endpoint_name} to the URL before signing.
            custom_headers (Optional[Dict[str, str]]): Additional query parameters to include
                in the presigned URL before signing (e.g., {"abc": "pqr"}).
            expires (int): Seconds until URL expires (default: 300, max: 300).

        Returns:
            str: Presigned WebSocket URL with query string parameters including:
                - Original query params (qualifier, custom_headers)
                - SigV4 auth params (X-Amz-Algorithm, X-Amz-Credential, etc.)

        Raises:
            ValueError: If expires exceeds maximum (300 seconds).
            RuntimeError: If URL generation fails or no credentials found.

        Example:
            >>> client = AgentCoreRuntimeClient('us-west-2')
            >>> presigned_url = client.generate_presigned_url(
            ...     runtime_arn='arn:aws:bedrock-agentcore:us-west-2:123:runtime/my-runtime',
            ...     endpoint_name='DEFAULT',
            ...     custom_headers={'abc': 'pqr'},
            ...     expires=300
            ... )
        """
        self.logger.info("Generating presigned WebSocket URL...")

        # Validate expires parameter
        if expires > MAX_PRESIGNED_URL_TIMEOUT:
            raise ValueError(f"Expiry timeout cannot exceed {MAX_PRESIGNED_URL_TIMEOUT} seconds, got {expires}")

        # Validate ARN
        self._parse_runtime_arn(runtime_arn)

        # Auto-generate session ID if not provided
        if not session_id:
            session_id = str(uuid.uuid4())
            self.logger.debug("Auto-generated session ID: %s", session_id)

        # Add session_id to custom_headers (which become query params)
        if custom_headers is None:
            custom_headers = {}
        custom_headers["X-Amzn-Bedrock-AgentCore-Runtime-Session-Id"] = session_id

        # Build WebSocket URL with query parameters
        ws_url = self._build_websocket_url(runtime_arn, endpoint_name, custom_headers)

        # Convert wss:// to https:// for signing
        https_url = ws_url.replace("wss://", "https://")

        # Parse URL
        url = urlparse(https_url)

        # Get AWS credentials
        credentials = self.session.get_credentials()
        if not credentials:
            raise RuntimeError("No AWS credentials found")

        frozen_credentials = credentials.get_frozen_credentials()

        # Create the request to sign
        request = AWSRequest(method="GET", url=https_url, headers={"host": url.hostname})

        # Sign the request with SigV4QueryAuth
        signer = SigV4QueryAuth(
            credentials=frozen_credentials,
            service_name="bedrock-agentcore",
            region_name=self.region,
            expires=expires,
        )
        signer.add_auth(request)

        if not request.url:
            raise RuntimeError("Failed to generate presigned URL")

        # Convert back to wss:// for WebSocket connection
        presigned_url = request.url.replace("https://", "wss://")

        self.logger.info("✓ Presigned URL generated (expires in %s seconds, Session: %s)", expires, session_id)
        return presigned_url

    def generate_ws_connection_oauth(
        self,
        runtime_arn: str,
        bearer_token: str,
        session_id: Optional[str] = None,
        endpoint_name: Optional[str] = None,
    ) -> Tuple[str, Dict[str, str]]:
        """Generate WebSocket URL and OAuth headers for runtime connection.

        This method uses OAuth bearer token authentication instead of AWS SigV4.
        Suitable for scenarios where OAuth tokens are used for authentication.

        Args:
            runtime_arn (str): Full runtime ARN
                (e.g., 'arn:aws:bedrock-agentcore:us-west-2:123:runtime/my-runtime-abc')
            bearer_token (str): OAuth bearer token for authentication.
            session_id (Optional[str]): Session ID to use. If None, auto-generates one.
            endpoint_name (Optional[str]): Endpoint name to use as 'qualifier' query parameter.
                If provided, adds ?qualifier={endpoint_name} to the URL.

        Returns:
            Tuple[str, Dict[str, str]]: A tuple containing:
                - WebSocket URL (wss://...) with query parameters
                - Headers dictionary with OAuth authentication

        Raises:
            ValueError: If runtime_arn format is invalid or bearer_token is empty.

        Example:
            >>> client = AgentCoreRuntimeClient('us-west-2')
            >>> ws_url, headers = client.generate_ws_connection_oauth(
            ...     runtime_arn='arn:aws:bedrock-agentcore:us-west-2:123:runtime/my-runtime',
            ...     bearer_token='eyJhbGciOiJSUzI1NiIsInR5cCI6IkpXVCJ9...',
            ...     endpoint_name='DEFAULT'
            ... )
        """
        self.logger.info("Generating WebSocket connection with OAuth authentication...")

        # Validate inputs
        if not bearer_token:
            raise ValueError("Bearer token cannot be empty")

        # Validate ARN
        self._parse_runtime_arn(runtime_arn)

        # Auto-generate session ID if not provided
        if not session_id:
            session_id = str(uuid.uuid4())
            self.logger.debug("Auto-generated session ID: %s", session_id)

        # Build WebSocket URL
        ws_url = self._build_websocket_url(runtime_arn, endpoint_name)

        # Convert wss:// to https:// to get host
        https_url = ws_url.replace("wss://", "https://")
        parsed = urlparse(https_url)

        # Generate WebSocket key
        ws_key = base64.b64encode(secrets.token_bytes(16)).decode()

        # Build OAuth headers
        headers = {
            "Authorization": f"Bearer {bearer_token}",
            "X-Amzn-Bedrock-AgentCore-Runtime-Session-Id": session_id,
            "Host": parsed.netloc,
            "Connection": "Upgrade",
            "Upgrade": "websocket",
            "Sec-WebSocket-Key": ws_key,
            "Sec-WebSocket-Version": "13",
            "User-Agent": "OAuth-WebSocket-Client/1.0",
        }

        self.logger.info("✓ OAuth WebSocket connection credentials generated (Session: %s)", session_id)
        self.logger.debug("Bearer token length: %d characters", len(bearer_token))

        return ws_url, headers



"""Tests for framework-agnostic ADOT models and builders."""

from unittest.mock import Mock

import pytest

from bedrock_agentcore.evaluation.span_to_adot_serializer.adot_models import (
    ADOTDocumentBuilder,
    ConversationTurn,
    ResourceInfo,
    SpanMetadata,
    SpanParser,
    ToolExecution,
)

# ==============================================================================
# Fixtures
# ==============================================================================


@pytest.fixture
def mock_span_context():
    """Create a mock span context."""
    context = Mock()
    context.trace_id = 0x1234567890ABCDEF1234567890ABCDEF
    context.span_id = 0x1234567890ABCDEF
    context.trace_flags = 1
    return context


@pytest.fixture
def mock_resource():
    """Create a mock resource."""
    resource = Mock()
    resource.attributes = {"service.name": "test-service"}
    return resource


@pytest.fixture
def mock_instrumentation_scope():
    """Create a mock instrumentation scope."""
    scope = Mock()
    scope.name = "strands.agent"
    scope.version = "1.0.0"
    return scope


@pytest.fixture
def mock_status():
    """Create a mock status."""
    status = Mock()
    status.status_code = Mock()
    status.status_code.__str__ = Mock(return_value="StatusCode.OK")
    return status


@pytest.fixture
def mock_span(mock_span_context, mock_resource, mock_instrumentation_scope, mock_status):
    """Create a mock OTel span."""
    span = Mock()
    span.context = mock_span_context
    span.resource = mock_resource
    span.instrumentation_scope = mock_instrumentation_scope
    span.status = mock_status
    span.parent = None
    span.name = "test-span"
    span.start_time = 1000000000
    span.end_time = 2000000000
    span.kind = Mock()
    span.kind.__str__ = Mock(return_value="SpanKind.INTERNAL")
    span.attributes = {"gen_ai.operation.name": "chat"}
    span.events = []
    return span


@pytest.fixture
def span_metadata():
    """Create test SpanMetadata."""
    return SpanMetadata(
        trace_id="1234567890abcdef1234567890abcdef",
        span_id="1234567890abcdef",
        parent_span_id=None,
        name="test-span",
        start_time=1000000000,
        end_time=2000000000,
        duration=1000000000,
        kind="INTERNAL",
        flags=1,
        status_code="OK",
    )


@pytest.fixture
def resource_info():
    """Create test ResourceInfo."""
    return ResourceInfo(
        resource_attributes={"service.name": "test-service"},
        scope_name="strands.agent",
        scope_version="1.0.0",
    )


# ==============================================================================
# Domain Model Tests
# ==============================================================================


class TestSpanMetadata:
    """Test SpanMetadata dataclass."""

    def test_creation(self):
        """Test SpanMetadata creation."""
        metadata = SpanMetadata(
            trace_id="abc123",
            span_id="def456",
            parent_span_id="parent123",
            name="test",
            start_time=1000,
            end_time=2000,
            duration=1000,
            kind="INTERNAL",
            flags=1,
            status_code="OK",
        )
        assert metadata.trace_id == "abc123"
        assert metadata.span_id == "def456"
        assert metadata.parent_span_id == "parent123"
        assert metadata.status_code == "OK"

    def test_optional_parent(self):
        """Test SpanMetadata with no parent."""
        metadata = SpanMetadata(
            trace_id="abc",
            span_id="def",
            parent_span_id=None,
            name="test",
            start_time=0,
            end_time=0,
            duration=0,
            kind="INTERNAL",
            flags=0,
            status_code="UNSET",
        )
        assert metadata.parent_span_id is None


class TestResourceInfo:
    """Test ResourceInfo dataclass."""

    def test_creation(self):
        """Test ResourceInfo creation."""
        info = ResourceInfo(
            resource_attributes={"service.name": "test"},
            scope_name="test.scope",
            scope_version="1.0.0",
        )
        assert info.resource_attributes == {"service.name": "test"}
        assert info.scope_name == "test.scope"
        assert info.scope_version == "1.0.0"


class TestConversationTurn:
    """Test ConversationTurn dataclass."""

    def test_creation(self):
        """Test ConversationTurn creation."""
        turn = ConversationTurn(
            user_message="Hello",
            assistant_messages=[{"content": {"message": "Hi"}, "role": "assistant"}],
            tool_results=["result1"],
        )
        assert turn.user_message == "Hello"
        assert len(turn.assistant_messages) == 1
        assert len(turn.tool_results) == 1


class TestToolExecution:
    """Test ToolExecution dataclass."""

    def test_creation(self):
        """Test ToolExecution creation."""
        tool = ToolExecution(
            tool_input='{"arg": "value"}',
            tool_output="result",
            tool_id="tool-123",
        )
        assert tool.tool_input == '{"arg": "value"}'
        assert tool.tool_output == "result"
        assert tool.tool_id == "tool-123"


# ==============================================================================
# Base Extraction Tests
# ==============================================================================


class TestSpanParser:
    """Test SpanParser class."""

    def test_extract_metadata(self, mock_span):
        """Test extracting metadata from span."""
        metadata = SpanParser.extract_metadata(mock_span)

        assert metadata.trace_id == "1234567890abcdef1234567890abcdef"
        assert metadata.span_id == "1234567890abcdef"
        assert metadata.parent_span_id is None
        assert metadata.name == "test-span"
        assert metadata.start_time == 1000000000
        assert metadata.end_time == 2000000000
        assert metadata.duration == 1000000000
        assert metadata.kind == "INTERNAL"
        assert metadata.flags == 1

    def test_extract_metadata_with_parent(self, mock_span):
        """Test extracting metadata from span with parent."""
        parent = Mock()
        parent.span_id = 0xFEDCBA0987654321
        mock_span.parent = parent

        metadata = SpanParser.extract_metadata(mock_span)

        assert metadata.parent_span_id == "fedcba0987654321"

    def test_extract_metadata_missing_context(self):
        """Test extracting metadata from span without context."""
        span = Mock()
        span.context = None
        span.name = "bad-span"

        with pytest.raises(ValueError, match="missing required context"):
            SpanParser.extract_metadata(span)

    def test_extract_resource_info(self, mock_span):
        """Test extracting resource info from span."""
        info = SpanParser.extract_resource_info(mock_span)

        assert info.resource_attributes == {"service.name": "test-service"}
        assert info.scope_name == "strands.agent"
        assert info.scope_version == "1.0.0"

    def test_extract_resource_info_missing_resource(self):
        """Test extracting resource info when resource is missing."""
        span = Mock()
        span.resource = None
        span.instrumentation_scope = None

        info = SpanParser.extract_resource_info(span)

        assert info.resource_attributes == {}
        assert info.scope_name == ""
        assert info.scope_version == ""

    def test_get_span_attributes(self, mock_span):
        """Test getting span attributes."""
        attrs = SpanParser.get_span_attributes(mock_span)

        assert attrs == {"gen_ai.operation.name": "chat"}

    def test_get_span_attributes_empty(self):
        """Test getting span attributes when empty."""
        span = Mock()
        span.attributes = None

        attrs = SpanParser.get_span_attributes(span)

        assert attrs == {}


# ==============================================================================
# ADOT Builder Tests
# ==============================================================================


class TestADOTDocumentBuilder:
    """Test ADOTDocumentBuilder class."""

    def test_build_span_document(self, span_metadata, resource_info):
        """Test building span document."""
        attributes = {"test.attr": "value"}

        doc = ADOTDocumentBuilder.build_span_document(span_metadata, resource_info, attributes)

        assert doc["traceId"] == "1234567890abcdef1234567890abcdef"
        assert doc["spanId"] == "1234567890abcdef"
        assert doc["name"] == "test-span"
        assert doc["kind"] == "INTERNAL"
        assert doc["startTimeUnixNano"] == 1000000000
        assert doc["endTimeUnixNano"] == 2000000000
        assert doc["durationNano"] == 1000000000
        assert doc["attributes"] == {"test.attr": "value"}
        assert doc["status"]["code"] == "OK"
        assert doc["resource"]["attributes"] == {"service.name": "test-service"}
        assert doc["scope"]["name"] == "strands.agent"

    def test_build_conversation_log_record(self, span_metadata, resource_info):
        """Test building conversation log record."""
        conversation = ConversationTurn(
            user_message="Hello",
            assistant_messages=[{"content": {"message": "Hi"}, "role": "assistant"}],
            tool_results=[],
        )

        doc = ADOTDocumentBuilder.build_conversation_log_record(conversation, span_metadata, resource_info)

        assert doc["traceId"] == "1234567890abcdef1234567890abcdef"
        assert doc["spanId"] == "1234567890abcdef"
        assert doc["severityNumber"] == 9
        assert doc["body"]["input"]["messages"][0]["content"]["content"] == "Hello"
        assert doc["body"]["output"]["messages"][0]["content"]["message"] == "Hi"

    def test_build_conversation_log_record_with_tool_results(self, span_metadata, resource_info):
        """Test building conversation log record with tool results."""
        conversation = ConversationTurn(
            user_message="Calculate",
            assistant_messages=[{"content": {"message": "4"}, "role": "assistant"}],
            tool_results=["4"],
        )

        doc = ADOTDocumentBuilder.build_conversation_log_record(conversation, span_metadata, resource_info)

        # Tool result attached to first assistant message
        assert doc["body"]["output"]["messages"][0]["content"]["tool.result"] == "4"

    def test_build_tool_log_record(self, span_metadata, resource_info):
        """Test building tool log record."""
        tool_exec = ToolExecution(
            tool_input='{"x": 1}',
            tool_output="result",
            tool_id="tool-123",
        )

        doc = ADOTDocumentBuilder.build_tool_log_record(tool_exec, span_metadata, resource_info)

        assert doc["traceId"] == "1234567890abcdef1234567890abcdef"
        assert doc["body"]["input"]["messages"][0]["content"]["content"] == '{"x": 1}'
        assert doc["body"]["input"]["messages"][0]["content"]["id"] == "tool-123"
        assert doc["body"]["output"]["messages"][0]["content"]["message"] == "result"
